import io
import json
import os
import shutil
import sys
import tempfile
from collections.abc import Generator
from contextlib import contextmanager
from email.message import EmailMessage
from email.parser import BytesParser
from functools import cache
from io import BytesIO
from operator import attrgetter
from pathlib import Path
from platform import python_version
from typing import TYPE_CHECKING, Any, BinaryIO, cast
from urllib.parse import urlparse
from zipfile import ZipFile

import requests
from build import ConfigSettingsType
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.version import Version
from resolvelib import BaseReporter, Resolver
from resolvelib.providers import AbstractProvider
from unearth.evaluator import TargetPython
from unearth.finder import PackageFinder

from pyodide_build import build_env
from pyodide_build.common import repack_zip_archive
from pyodide_build.logger import logger
from pyodide_build.out_of_tree import build
from pyodide_build.spec import _BuildSpecExports

_PYPI_INDEX = ["https://pypi.org/simple/"]
_PYPI_TRUSTED_HOSTS = ["pypi.org"]


@contextmanager
def stream_redirected(to=os.devnull, stream=None):
    """
    Context manager to redirect stdout or stderr. It does it with filenos and things rather than
    just changing sys.stdout, so that output of subprocesses is also redirected.
    """
    if stream is None:
        stream = sys.stdout
    try:
        if not hasattr(stream, "fileno"):
            yield
            return
        stream_fd = stream.fileno()
    except io.UnsupportedOperation:
        # in case we're already capturing to something that isn't really a file
        # e.g. in pytest
        yield
        return
    if isinstance(to, str):
        to = open(to, "w")
    with os.fdopen(os.dup(stream_fd), "wb") as copied:
        stream.flush()
        os.dup2(to.fileno(), stream_fd)  # $ exec >&to
        try:
            yield stream  # allow code to be run with the redirected stream
        finally:
            # restore stream to its previous value
            # NOTE: dup2 makes stream_fd inheritable unconditionally
            stream.flush()
            os.dup2(copied.fileno(), stream_fd)  # $ exec >&copied
            to = None


def get_built_wheel(url, isolation=True, skip_dependency_check=False):
    return _get_built_wheel_internal(url, isolation, skip_dependency_check)["path"]


@cache
def _get_built_wheel_internal(url, isolation=True, skip_dependency_check=False):
    parsed_url = urlparse(url)
    gz_name = Path(parsed_url.path).name

    cache_entry: dict[str, Any] = {}
    build_dir = tempfile.TemporaryDirectory()
    build_path = Path(build_dir.name)

    cache_entry["build_dir"] = build_dir
    with tempfile.NamedTemporaryFile(suffix=".tar.gz", delete=False) as f:
        data = requests.get(url).content
        f.write(data)
        f.close()
        shutil.unpack_archive(f.name, build_path)
        os.unlink(f.name)
        files = list(build_path.iterdir())
        if len(files) == 1 and files[0].is_dir():
            source_path = build_path / files[0]
        else:
            source_path = build_path
    logger.info("Building wheel for %s...", gz_name)
    with (
        tempfile.NamedTemporaryFile(mode="w+") as logfile,
        stream_redirected(to=logfile, stream=sys.stdout),
        stream_redirected(to=logfile, stream=sys.stderr),
    ):
        try:
            wheel_path = build.run(
                source_path,
                build_path / "dist",
                PyPIProvider.BUILD_EXPORTS,
                PyPIProvider.BUILD_FLAGS,
                isolation=isolation,
                skip_dependency_check=skip_dependency_check,
            )
        except BaseException as e:
            logger.error(" Failed\n Error is:")
            logfile.seek(0)
            logger.stderr(logfile.read())
            raise e

    logger.success("Success")

    cache_entry["path"] = wheel_path
    return cache_entry


class Candidate:
    def __init__(self, name, version, url=None, extras=None):
        self.name = canonicalize_name(name)
        self.version = version
        self.url = url
        self.extras = extras

        self._metadata = None
        self._dependencies = None

    def __repr__(self):
        if not self.extras:
            return f"<{self.name}=={self.version}>"
        return f"<{self.name}[{','.join(self.extras)}]=={self.version}>"

    @property
    def metadata(self):
        if self._metadata is None:
            self._metadata = get_metadata_for_wheel(self.url)
        return self._metadata

    @property
    def requires_python(self):
        return self.metadata.get("Requires-Python")

    def _get_dependencies(self):
        deps = self.metadata.get_all("Requires-Dist", [])
        extras = self.extras if self.extras else [""]

        for d in deps:
            r = Requirement(d)
            if r.marker is None:
                yield r
            else:
                for e in extras:
                    if r.marker.evaluate({"extra": e}):
                        yield r
                        break

    @property
    def dependencies(self):
        if self._dependencies is None:
            self._dependencies = list(self._get_dependencies())
        return self._dependencies


if TYPE_CHECKING:
    APBase = AbstractProvider[Requirement, Candidate, str]
else:
    APBase = AbstractProvider

PYTHON_VERSION = Version(python_version())


def get_target_python():
    PYMAJOR = build_env.get_pyversion_major()
    PYMINOR = build_env.get_pyversion_minor()
    tp = TargetPython(
        py_ver=(int(PYMAJOR), int(PYMINOR)),
        platforms=[build_env.platform()],
        abis=[f"cp{PYMAJOR}{PYMINOR}"],
    )
    return tp


def get_project_from_pypi(package_name, extras):
    """Return candidates created from the project name and extras."""
    pf = PackageFinder(
        index_urls=_PYPI_INDEX,
        trusted_hosts=_PYPI_TRUSTED_HOSTS,
        target_python=get_target_python(),
    )
    matches = pf.find_all_packages(package_name)
    for i in matches:
        # TODO: ignore sourcedists if wheel for same version exists
        yield Candidate(i.name, i.version, url=i.link.url, extras=extras)


def download_or_build_wheel(
    url: str,
    target_directory: Path,
    compression_level: int = 6,
    isolation: bool = True,
    skip_dependency_check: bool = False,
) -> None:
    parsed_url = urlparse(url)
    if parsed_url.path.endswith("gz"):
        wheel_file = get_built_wheel(url, isolation, skip_dependency_check)
        shutil.copy(wheel_file, target_directory)
        wheel_path = target_directory / wheel_file.name
    elif parsed_url.path.endswith(".whl"):
        wheel_path = target_directory / Path(parsed_url.path).name
        with open(wheel_path, "wb") as f:
            f.write(requests.get(url).content)

    repack_zip_archive(wheel_path, compression_level=compression_level)


def get_metadata_for_wheel(url):
    parsed_url = urlparse(url)
    if parsed_url.path.endswith("gz"):
        wheel_file = get_built_wheel(url)
        wheel_stream: BinaryIO = open(wheel_file, "rb")
    elif parsed_url.path.endswith(".whl"):
        data = requests.get(url).content
        wheel_stream = BytesIO(data)
    else:
        raise RuntimeError(f"Distributions of this type are unsupported:{url}")
    with ZipFile(wheel_stream) as z:
        for n in z.namelist():
            if n.endswith(".dist-info/METADATA"):
                p = BytesParser()
                return p.parse(cast(BinaryIO, z.open(n)), headersonly=True)

    # If we didn't find the metadata, return an empty dict
    return EmailMessage()


class PyPIProvider(APBase):
    BUILD_FLAGS: ConfigSettingsType = {}
    BUILD_SKIP: list[str] = []
    BUILD_EXPORTS: _BuildSpecExports = []

    def __init__(self, build_dependencies: bool):
        self.build_dependencies = build_dependencies

    def identify(self, requirement_or_candidate):
        base = canonicalize_name(requirement_or_candidate.name)
        return base

    def get_extras_for(self, requirement_or_candidate):
        # Extras is a set, which is not hashable
        return tuple(sorted(requirement_or_candidate.extras))

    def get_base_requirement(self, candidate):
        return Requirement(f"{candidate.name}=={candidate.version}")

    def get_preference(
        self, identifier, resolutions, candidates, information, backtrack_causes
    ):
        return sum(1 for _ in candidates[identifier])

    def find_matches(self, identifier, requirements, incompatibilities):
        requirements = list(requirements[identifier])

        extra_requirements = {}
        for r in requirements:
            extra_requirements[tuple(r.extras)] = 1

        bad_versions = {c.version for c in incompatibilities[identifier]}

        # Need to pass the extras to the search, so they
        # are added to the candidate at creation - we
        # treat candidates as immutable once created.
        for extra_tuple in extra_requirements.keys():
            extras = set(extra_tuple)

            candidates = (
                candidate
                for candidate in get_project_from_pypi(identifier, extras)
                if candidate.version not in bad_versions
                and all(candidate.version in r.specifier for r in requirements)
            )

        return sorted(candidates, key=attrgetter("version"), reverse=True)

    def is_satisfied_by(self, requirement, candidate):
        if canonicalize_name(requirement.name) != candidate.name:
            return False
        return candidate.version in requirement.specifier

    def get_dependencies(self, candidate):
        deps = []
        if self.build_dependencies:
            for d in candidate.dependencies:
                if d.name not in PyPIProvider.BUILD_SKIP:
                    deps.append(d)
        if candidate.extras:
            # add the base package as a dependency too, so we can avoid conflicts between same package
            # but with different extras
            req = self.get_base_requirement(candidate)
            deps.append(req)
        return deps


def _get_json_package_list(fname: Path) -> Generator[str, None, None]:
    json_data = json.load(fname.open())
    if "packages" in json_data:
        # pyodide-lock.json format
        yield from json_data["packages"].keys()
    else:
        # jupyterlite all.json format
        for k in json_data.keys():
            if "releases" in json_data[k]:
                yield k


def _parse_skip_list(skip_dependency: list[str]) -> None:
    PyPIProvider.BUILD_SKIP = []
    for skip in skip_dependency:
        split_deps = skip.split(",")
        for dep in split_deps:
            if dep.endswith(".json"):
                # a pyodide json file
                # or a jupyterlite json file
                # skip all packages in it
                PyPIProvider.BUILD_SKIP.extend(_get_json_package_list(Path(dep)))
            else:
                PyPIProvider.BUILD_SKIP.append(dep)


def _resolve_and_build(
    deps: list[str],
    target_folder: Path,
    build_dependencies: bool,
    extras: list[str],
    output_lockfile: str | None,
    isolation: bool = True,
    skip_dependency_check: bool = False,
    compression_level: int = 6,
) -> None:
    requirements = []

    target_env = {
        "python_version": build_env.get_pyversion_major_minor(),
        "sys_platform": build_env.platform().split("_")[0],
        "extra": ",".join(extras),
    }

    for d in deps:
        r = Requirement(d)
        if (r.name not in PyPIProvider.BUILD_SKIP) and (
            (not r.marker) or r.marker.evaluate(target_env)
        ):
            requirements.append(r)

    # Create the (reusable) resolver.
    provider = PyPIProvider(build_dependencies=build_dependencies)
    reporter = BaseReporter()
    resolver: Resolver[Requirement, Candidate, str] = Resolver(provider, reporter)

    # Kick off the resolution process, and get the final result.
    result = resolver.resolve(requirements)
    target_folder.mkdir(parents=True, exist_ok=True)
    version_file = None
    if output_lockfile is not None and len(output_lockfile) > 0:
        version_file = open(output_lockfile, "w")
    for x in result.mapping.values():
        download_or_build_wheel(x.url, target_folder, compression_level)
        if len(x.extras) > 0:
            extratxt = "[" + ",".join(x.extras) + "]"
        else:
            extratxt = ""
        if version_file:
            version_file.write(f"{x.name}{extratxt}=={x.version}\n")
    if version_file:
        version_file.close()


def build_wheels_from_pypi_requirements(
    reqs: list[str],
    target_folder: Path,
    build_dependencies: bool,
    skip_dependency: list[str],
    exports: _BuildSpecExports,
    config_settings: ConfigSettingsType,
    isolation: bool = True,
    skip_dependency_check: bool = False,
    output_lockfile: str | None = None,
    compression_level: int = 6,
) -> None:
    """
    Given a list of package requirements, build or fetch them. If build_dependencies is true, then
    package dependencies will be built or fetched also.
    """
    _parse_skip_list(skip_dependency)
    PyPIProvider.BUILD_EXPORTS = exports
    PyPIProvider.BUILD_FLAGS = config_settings
    _resolve_and_build(
        reqs,
        target_folder,
        build_dependencies,
        extras=[],
        output_lockfile=output_lockfile,
        isolation=isolation,
        skip_dependency_check=skip_dependency_check,
        compression_level=compression_level,
    )


def build_dependencies_for_wheel(
    wheel: Path,
    extras: list[str],
    skip_dependency: list[str],
    exports: _BuildSpecExports,
    config_settings: ConfigSettingsType,
    isolation: bool = True,
    skip_dependency_check: bool = False,
    output_lockfile: str | None = None,
    compression_level: int = 6,
) -> None:
    """Extract dependencies from this wheel and build pypi dependencies
    for each one in ./dist/

    n.b. because dependency resolution may need to backtrack, this
    is potentially quite slow in the case that one needs to build an
    sdist in order to discover dependencies of a candidate sub-dependency.
    """
    metadata = None
    _parse_skip_list(skip_dependency)

    PyPIProvider.BUILD_EXPORTS = exports
    PyPIProvider.BUILD_FLAGS = config_settings
    with ZipFile(wheel) as z:
        for n in z.namelist():
            if n.endswith(".dist-info/METADATA"):
                p = BytesParser()
                metadata = p.parse(cast(BinaryIO, z.open(n)), headersonly=True)
    if metadata is None:
        raise RuntimeError(f"Can't find package metadata in {wheel}")

    deps: list[str] = metadata.get_all("Requires-Dist", [])
    metadata.get("version")
    _resolve_and_build(
        deps,
        wheel.parent,
        build_dependencies=True,
        extras=extras,
        output_lockfile=output_lockfile,
        isolation=isolation,
        skip_dependency_check=skip_dependency_check,
        compression_level=compression_level,
    )
    # add the current wheel to the package-versions.txt
    if output_lockfile is not None and len(output_lockfile) > 0:
        with open(output_lockfile, "a") as version_txt:
            name = metadata.get("Name")
            version = metadata.get("Version")
            if extras:
                extratxt = "[" + ",".join(extras) + "]"
            else:
                extratxt = ""
            version_txt.write(f"{name}{extratxt}=={version}\n")


def fetch_pypi_package(package_spec: str, destdir: Path) -> Path:
    pf = PackageFinder(
        index_urls=_PYPI_INDEX,
        trusted_hosts=_PYPI_TRUSTED_HOSTS,
        target_python=get_target_python(),
    )
    match = pf.find_best_match(package_spec)
    if match.best is None:
        if len(match.candidates) != 0:
            error = f"""Can't find version matching {package_spec}
versions found:
"""
            for c in match.candidates:
                error += "  " + str(c.version) + "\t"
            raise RuntimeError(error)
        else:
            raise RuntimeError(f"Can't find package: {package_spec}")
    with tempfile.TemporaryDirectory() as download_dir:
        return pf.download_and_unpack(
            link=match.best.link, location=destdir, download_dir=download_dir
        )



import shutil
import subprocess
import sys
import textwrap
from pathlib import Path
from typing import Any

from pyodide_build.build_env import get_build_flag, get_pyodide_root, in_xbuildenv
from pyodide_build.common import exit_with_stdio
from pyodide_build.logger import logger

# A subset of supported virtualenv options that make sense in Pyodide's context.
# Our aim will not be to support all of them, and some of them will never be in
# the list, for example, --no-pip and so on. We provide these on a best-effort
# basis as they should work and are easy to test.
SUPPORTED_VIRTUALENV_OPTIONS = [
    "--clear",
    "--no-clear",
    "--no-vcs-ignore",
    # "--copies", "--always-copy", FIXME: node fails to invoke Pyodide
    # "--symlink-app-data", FIXME: node fails to invoke Pyodide
    "--no-download",
    "--never-download",
    "--download",
    "--extra-search-dir",
    "--pip",
    "--setuptools",
    "--no-setuptools",
    "--no-periodic-update",
]


def check_result(result: subprocess.CompletedProcess[str], msg: str) -> None:
    """Abort if the process returns a nonzero error code"""
    if result.returncode != 0:
        logger.error(msg)
        exit_with_stdio(result)


def dedent(s: str) -> str:
    return textwrap.dedent(s).strip() + "\n"


def get_pyversion() -> str:
    return f"{sys.version_info.major}.{sys.version_info.minor}"


def check_host_python_version(session: Any) -> None:
    pyodide_version = session.interpreter.version.partition(" ")[0].split(".")[:2]
    sys_version = [str(sys.version_info.major), str(sys.version_info.minor)]
    if pyodide_version == sys_version:
        return
    pyodide_version_fmt = ".".join(pyodide_version)
    sys_version_fmt = ".".join(sys_version)
    logger.stderr(
        f"Expected host Python version to be {pyodide_version_fmt} but got version {sys_version_fmt}"
    )
    sys.exit(1)


def pyodide_dist_dir() -> Path:
    return get_pyodide_root() / "dist"


def create_pip_conf(venv_root: Path) -> None:
    """Create pip.conf file in venv root

    This file adds a few options that will always be used by pip install.
    """
    if in_xbuildenv():
        # In the xbuildenv, we don't have the packages locally. We will include
        # in the xbuildenv a PEP 503 index for the vendored Pyodide packages
        # https://peps.python.org/pep-0503/
        repo = f"extra-index-url=file:{get_pyodide_root() / 'package_index'}"
    else:
        # In the Pyodide development environment, the Pyodide dist directory
        # should contain the needed wheels. find-links
        repo = f"find-links={pyodide_dist_dir()}"

    # Prevent attempts to install binary wheels from source.
    # Maybe some day we can convince pip to invoke `pyodide build` as the build
    # front end for wheels...
    (venv_root / "pip.conf").write_text(
        dedent(
            f"""
            [install]
            only-binary=:all:
            {repo}
            """
        )
    )


def get_pip_monkeypatch(venv_bin: Path) -> str:
    """Monkey patch pip's environment to show info about Pyodide's environment.

    The code returned is injected at the beginning of the pip script.
    """
    result = subprocess.run(
        [
            venv_bin / "python",
            "-c",
            dedent(
                """
                import os, sys, sysconfig, platform
                print([
                    os.name,
                    sys.platform,
                    platform.system(),
                    sys.implementation._multiarch,
                    sysconfig.get_platform()
                ])
                """
            ),
        ],
        capture_output=True,
        encoding="utf8",
        check=False,
    )
    check_result(result, "ERROR: failed to invoke Pyodide")
    platform_data = result.stdout
    sysconfigdata_dir = Path(get_build_flag("TARGETINSTALLDIR")) / "sysconfigdata"
    return dedent(
        """\
        import os
        import platform
        import sys
        """
        # when pip installs an executable it uses sys.executable to create the
        # shebang for the installed executable. The shebang for pip points to
        # python-host but we want the shebang of the executable that we install
        # to point to Pyodide python. We monkeypatch distlib.scripts.get_executable
        # to return the value with the host suffix removed.
        """
        from pip._vendor.distlib import scripts
        EXECUTABLE_SUFFIX = "-host-link"
        def get_executable():
            if not sys.executable.endswith(EXECUTABLE_SUFFIX):
                raise RuntimeError(f'Internal Pyodide error: expected sys.executable="{sys.executable}" to end with "{EXECUTABLE_SUFFIX}"')
            return sys.executable.removesuffix(EXECUTABLE_SUFFIX)

        scripts.get_executable = get_executable

        from pip._vendor.packaging import tags
        orig_platform_tags = tags.platform_tags
        """
        # TODO: Remove the following monkeypatch when we merge and pull in
        # https://github.com/pypa/packaging/pull/804
        """
        def _emscripten_platforms():
            pyodide_abi_version = sysconfig.get_config_var("PYODIDE_ABI_VERSION")
            if pyodide_abi_version:
                yield f"pyodide_{pyodide_abi_version}_wasm32"
            yield from tags._generic_platforms()

        def platform_tags():
            if platform.system() == "Emscripten":
                yield from _emscripten_platforms()
                return
            return orig_platform_tags()

        tags.platform_tags = platform_tags
        """
        f"""
        os_name, sys_platform, platform_system, multiarch, host_platform = {platform_data}
        os.name = os_name
        sys.platform = sys_platform
        sys.platlibdir = "lib"
        sys.implementation._multiarch = multiarch
        platform.system = lambda: platform_system
        platform.machine = lambda: "wasm32"
        os.environ["_PYTHON_HOST_PLATFORM"] = host_platform
        os.environ["_PYTHON_SYSCONFIGDATA_NAME"] = f'_sysconfigdata_{{sys.abiflags}}_{{sys.platform}}_{{sys.implementation._multiarch}}'
        sys.path.append("{sysconfigdata_dir}")
        import sysconfig
        sysconfig._init_config_vars()
        del os.environ["_PYTHON_SYSCONFIGDATA_NAME"]
        """
        # Handle pip updates.
        #
        # The pip executable should be a symlink to pip_patched. If it is not a
        # link, or it is a symlink to something else, pip has been updated. We
        # have to restore the correct value of pip. Iterate through all of the
        # pip variants in the folder and remove them and replace with a symlink
        # to pip_patched.
        """
        from pathlib import Path

        file_path = Path(__file__)


        def pip_is_okay():
            try:
                return file_path.readlink() == file_path.with_name("pip_patched")
            except OSError as e:
                if e.strerror != "Invalid argument":
                    raise
            return False


        def maybe_repair_after_pip_update():
            if pip_is_okay():
                return

            venv_bin = file_path.parent
            pip_patched = venv_bin / "pip_patched"
            for pip in venv_bin.glob("pip*"):
                if pip == pip_patched:
                    continue
                pip.unlink(missing_ok=True)
                pip.symlink_to(venv_bin / "pip_patched")


        import atexit

        atexit.register(maybe_repair_after_pip_update)
        """
    )


def create_pip_script(venv_bin):
    """Create pip and write it into the virtualenv bin folder."""
    # pip needs to run in the host Python not in Pyodide, so we'll use the host
    # Python in the shebang. Use whichever Python was used to invoke
    # pyodide venv.
    host_python_path = venv_bin / f"python{get_pyversion()}-host"
    host_python_path_no_version = venv_bin / "python-host"
    pip_path = venv_bin / "pip_patched"
    python_host_link = venv_bin / "python-host-link"

    # To support the "--clear" and "--no-clear" args, we need to remove
    # the existing symlinks before creating new ones.
    host_python_path.unlink(missing_ok=True)
    host_python_path_no_version.unlink(missing_ok=True)
    python_host_link.unlink(missing_ok=True)
    for pip in venv_bin.glob("pip*"):
        if pip == pip_path:
            continue
        pip.unlink(missing_ok=True)
        pip.symlink_to(pip_path)

    # Weird hack to work around:
    # https://github.com/astral-sh/python-build-standalone/issues/380
    # If we resolve the symlink all the way, the python-host interpreter works
    # but won't install into our pyodide venv. If we don't resolve the symlink,
    # sys.prefix is calculated incorrectly. To ensure that we get the right
    # sys.prefix, we explicitly set it with the PYTHONHOME environment variable
    # and then call the symlink.
    python_host_link.symlink_to(sys.executable)
    pythonhome = Path(sys._base_executable).parents[1]
    host_python_path.write_text(
        dedent(
            f"""\
            #!/bin/sh
            exec env PYTHONHOME={pythonhome} {python_host_link} "$@"
            """
        )
    )
    host_python_path.chmod(0o777)
    host_python_path_no_version.symlink_to(host_python_path)

    pip_path.write_text(
        # Other than the shebang and the monkey patch, this is exactly what
        # normal pip looks like.
        f"#!/usr/bin/env -S {host_python_path} -s\n"
        + get_pip_monkeypatch(venv_bin)
        + dedent(
            """
            import re
            import sys
            from pip._internal.cli.main import main
            if __name__ == '__main__':
                sys.argv[0] = re.sub(r'(-script\\.pyw|\\.exe)?$', '', sys.argv[0])
                sys.exit(main())
            """
        )
    )
    pip_path.chmod(0o777)


def create_pyodide_script(venv_bin: Path) -> None:
    """Write pyodide cli script into the virtualenv bin folder"""
    import os

    # Temporarily restore us to the environment that 'pyodide venv' was
    # invoked in
    PATH = os.environ["PATH"]
    PYODIDE_ROOT = os.environ["PYODIDE_ROOT"]

    original_pyodide_cli = shutil.which("pyodide")
    if original_pyodide_cli is None:
        raise RuntimeError("ERROR: pyodide cli not found")

    pyodide_path = venv_bin / "pyodide"
    pyodide_path.write_text(
        dedent(
            f"""
            #!/usr/bin/env bash
            PATH="{PATH}:$PATH" PYODIDE_ROOT='{PYODIDE_ROOT}' exec {original_pyodide_cli} "$@"
            """
        )
    )
    pyodide_path.chmod(0o777)


def install_stdlib(venv_bin: Path) -> None:
    """Install micropip and all unvendored stdlib modules"""
    # Micropip we could install with pip hypothetically, but because we use
    # `--extra-index-url` it would install the pypi version which we don't want.

    # Other stuff we need to load with loadPackage
    to_load = ["micropip"]
    result = subprocess.run(
        [
            venv_bin / "python",
            "-c",
            dedent(
                f"""
                from pyodide_js import loadPackage
                from pyodide_js._api import lockfile_packages
                from pyodide_js._api import lockfile_unvendored_stdlibs_and_test
                shared_libs = [pkgname for (pkgname,pkg) in lockfile_packages.object_entries() if getattr(pkg, "package_type", None) == "shared_library"]

                to_load = [*lockfile_unvendored_stdlibs_and_test, *shared_libs, *{to_load!r}]
                loadPackage(to_load);
                """
            ),
        ],
        capture_output=True,
        encoding="utf8",
        check=False,
    )
    check_result(result, "ERROR: failed to install unvendored stdlib modules")


def create_pyodide_venv(dest: Path, virtualenv_args: list[str] | None = None) -> None:
    """Create a Pyodide virtualenv and store it into dest"""
    logger.info("Creating Pyodide virtualenv at %s", dest)
    from virtualenv import session_via_cli

    interp_path = pyodide_dist_dir() / "python"

    cli_args = ["--python", str(interp_path)]

    if virtualenv_args:
        for arg in virtualenv_args:
            if arg.startswith("--"):
                # Check if the argument (or its prefix form) is supported.
                arg_name = arg.split("=")[0] if "=" in arg else arg
                if arg_name not in SUPPORTED_VIRTUALENV_OPTIONS:
                    msg = f"Unsupported virtualenv option: {arg_name}"
                    logger.warning(msg)

        cli_args.extend(virtualenv_args)

    session = session_via_cli(cli_args + [str(dest)])
    check_host_python_version(session)

    try:
        session.run()
        venv_root = Path(session.creator.dest).absolute()
        venv_bin = venv_root / "bin"

        logger.info("... Configuring virtualenv")
        create_pip_conf(venv_root)
        create_pip_script(venv_bin)
        create_pyodide_script(venv_bin)
        logger.info("... Installing standard library")
        install_stdlib(venv_bin)
    except (Exception, KeyboardInterrupt, SystemExit):
        shutil.rmtree(session.creator.dest)
        raise

    logger.success("Successfully created Pyodide virtual environment!")



import os
from pathlib import Path
from textwrap import dedent

from build import ConfigSettingsType

from pyodide_build import build_env, common, pypabuild
from pyodide_build.build_env import get_pyodide_root, wheel_platform
from pyodide_build.spec import _BuildSpecExports


def _create_ignore_files(directory: Path) -> None:
    directory.joinpath(".gitignore").write_text(
        dedent("""\
        # Created by pyodide-build
        *
    """),
        encoding="utf-8",
    )

    directory.joinpath(".hgignore").write_text(
        dedent("""\
        # Created by pyodide-build
        syntax: glob
        **/*
    """),
        encoding="utf-8",
    )


def _prepare_build_dir(build_dir: Path) -> None:
    # create a persistent build dir in the source dir
    build_dir.mkdir(exist_ok=True)
    # don't track the build dir in version control,
    # helps if building in a git/mercurial repo
    _create_ignore_files(build_dir)


def run(
    srcdir: Path,
    outdir: Path,
    exports: _BuildSpecExports,
    config_settings: ConfigSettingsType,
    isolation: bool = True,
    skip_dependency_check: bool = False,
) -> Path:
    outdir = outdir.resolve()
    cflags = build_env.get_build_flag("SIDE_MODULE_CFLAGS")
    cflags += f" {os.environ.get('CFLAGS', '')}"
    cxxflags = build_env.get_build_flag("SIDE_MODULE_CXXFLAGS")
    cxxflags += f" {os.environ.get('CXXFLAGS', '')}"
    ldflags = build_env.get_build_flag("SIDE_MODULE_LDFLAGS")
    ldflags += f" {os.environ.get('LDFLAGS', '')}"
    target_install_dir = os.environ.get(
        "TARGETINSTALLDIR", build_env.get_build_flag("TARGETINSTALLDIR")
    )
    build_env._create_constraints_file()
    env = os.environ.copy()
    env.update(build_env.get_build_environment_vars(get_pyodide_root()))

    build_dir = srcdir / ".pyodide_build"
    _prepare_build_dir(build_dir)

    build_env_ctx = pypabuild.get_build_env(
        env=env,
        pkgname="",
        cflags=cflags,
        cxxflags=cxxflags,
        ldflags=ldflags,
        target_install_dir=target_install_dir,
        exports=exports,
        build_dir=build_dir,
    )

    with build_env_ctx as env:
        built_wheel = pypabuild.build(
            srcdir,
            outdir,
            env,
            config_settings,
            isolation=isolation,
            skip_dependency_check=skip_dependency_check,
        )

    wheel_path = Path(built_wheel)
    if "emscripten" in wheel_path.name:
        # Retag platformed wheels to pyodide
        wheel_path = common.retag_wheel(wheel_path, wheel_platform())
    with common.modify_wheel(wheel_path) as wheel_dir:
        build_env.replace_so_abi_tags(wheel_dir)

    return wheel_path



# Create or update a package recipe skeleton,
# inspired from `conda skeleton` command.

import sys
from pathlib import Path

import typer

from pyodide_build import build_env
from pyodide_build.logger import logger
from pyodide_build.recipe import skeleton

app = typer.Typer()


@app.callback(no_args_is_help=True)
def callback() -> None:
    """Add a new package build recipe or update an existing recipe"""
    return


def get_recipe_dir(recipe_dir: str | None) -> Path:
    if recipe_dir:
        return Path(recipe_dir)
    cwd = Path.cwd()
    root = build_env.search_pyodide_root(curdir=cwd)

    if not root:
        root = cwd

    return root / "packages"


@app.command("enable")
def enable(
    names: list[str],
    recipe_dir: str | None = typer.Option(
        None,
        help="The directory containing the recipe of packages. "
        "If not specified, the default is ``<cwd>/packages``.",
    ),
):
    recipe_dir_ = get_recipe_dir(recipe_dir)
    status = 0
    for name in names:
        try:
            skeleton.enable_package(
                recipe_dir_,
                name,
            )
        except skeleton.MkpkgFailedException as e:
            status = -1
            logger.error("%s update failed: %s", name, e)
        except Exception:
            print(name)
            raise
    sys.exit(status)


@app.command("disable")
def disable(
    names: list[str],
    message: str = typer.Option(
        "", "--message", "-m", help="Comment to explain why it was disabled"
    ),
    recipe_dir: str | None = typer.Option(
        None,
        help="The directory containing the recipe of packages. "
        "If not specified, the default is ``<cwd>/packages``.",
    ),
) -> int:
    recipe_dir_ = get_recipe_dir(recipe_dir)
    status = 0
    for name in names:
        try:
            skeleton.disable_package(recipe_dir_, name, message)
        except skeleton.MkpkgFailedException as e:
            status = -1
            logger.error("%s update failed: %s", name, e)
        except Exception:
            print(name)
            raise
    sys.exit(status)


@app.command("pin")
def pin(
    names: list[str],
    message: str = typer.Option(
        "", "--message", "-m", help="Comment to explain why it was pinned"
    ),
    recipe_dir: str | None = typer.Option(
        None,
        help="The directory containing the recipe of packages. "
        "If not specified, the default is ``<cwd>/packages``.",
    ),
) -> int:
    recipe_dir_ = get_recipe_dir(recipe_dir)
    status = 0
    for name in names:
        try:
            skeleton.pin_package(recipe_dir_, name, message)
        except skeleton.MkpkgFailedException as e:
            status = -1
            logger.error("%s update failed: %s", name, e)
        except Exception:
            print(name)
            raise
    sys.exit(status)


@app.command("pypi")
def new_recipe_pypi(
    name: str,
    update: bool = typer.Option(
        False,
        "--update",
        "-u",
        help="Update an existing recipe instead of creating a new one.",
    ),
    update_patched: bool = typer.Option(
        False,
        "--update-patched",
        help="Force update the package even if it contains patches.",
    ),
    update_pinned: bool = typer.Option(
        False,
        "--update-pinned",
        help="Force update the package even if is pinned.",
    ),
    version: str = typer.Option(
        None,
        help="The version of the package, if not specified, latest version will be used.",
    ),
    source_format: str = typer.Option(
        None,
        help="Which source format is preferred. Options are wheel or sdist. "
        "If not specified, then either a wheel or an sdist will be used. ",
    ),
    recipe_dir: str | None = typer.Option(
        None,
        help="The directory containing the recipe of packages. "
        "If not specified, the default is ``<cwd>/packages``.",
    ),
    maintainer: str | None = typer.Option(
        None, "--maintainer", "-m", help="The github username to use as the maintainer"
    ),
    gh_maintainer: bool = typer.Option(
        False,
        "--gh-maintainer",
        "--ghm",
        help="Set the maintainer from 'gh auth status'. Requires gh cli.",
    ),
) -> None:
    """
    Create a new package recipe from PyPI or update an existing recipe.
    """

    # Determine the recipe directory. If it is specified by the user, we use that;
    # otherwise, we assume that the recipe directory is the ``packages`` directory
    # in the root of the Pyodide tree, without the need to initialize the
    # cross-build environment.
    #
    # It is unlikely that a user will run this command outside of the Pyodide
    # tree, so we do not need to initialize the environment at this stage.

    recipe_dir_ = get_recipe_dir(recipe_dir)

    try:
        if update or update_patched or update_pinned:
            action = "update"
            if maintainer or gh_maintainer:
                raise skeleton.MkpkgFailedException(
                    "--maintainer and --gh-maintainer are only currently supported when creating a new recipe."
                )
            skeleton.update_package(
                recipe_dir_,
                name,
                version=version,
                source_fmt=source_format,  # type: ignore[arg-type]
                update_patched=update_patched,
                update_pinned=update_pinned,
            )
        else:
            action = "create"
            if maintainer and gh_maintainer:
                raise skeleton.MkpkgFailedException(
                    "At most one of --maintainer and --gh-maintainer can use used."
                )
            if gh_maintainer:
                maintainer = skeleton.lookup_gh_username()
            skeleton.make_package(
                recipe_dir_,
                name,
                version,
                source_fmt=source_format,
                maintainer=maintainer,
            )  # type: ignore[arg-type]

    except skeleton.MkpkgFailedException as e:
        logger.error("Failed to %s %s: %s", action, name, e)
        sys.exit(1)
    except skeleton.MkpkgSkipped as e:
        logger.warning("%s %s skipped: %s", name, action, e)
    except Exception:
        print(name)
        raise



# Some reusable constants that are used at various sources in our code.
# TODO: collect and add more constants here.

BASE_IGNORED_REQUIREMENTS: list[str] = [
    # mesonpy installs patchelf in linux platform but we don't want it.
    "patchelf",
    "oldest-supported-numpy",
]



import functools
from collections.abc import Iterable
from pathlib import Path

from pyodide_build.logger import logger
from pyodide_build.recipe.spec import MetaConfig


@functools.lru_cache(maxsize=1)
def load_all_recipes(recipe_dir: Path) -> dict[str, MetaConfig]:
    """Load all package recipes from the recipe directory."""

    recipes_path = recipe_dir.glob("*/meta.yaml")

    recipes: dict[str, MetaConfig] = {}
    for recipe in recipes_path:
        try:
            config = MetaConfig.from_yaml(recipe)
            recipes[config.package.name] = config
        except Exception as e:
            raise ValueError(f"Could not parse {recipe}.") from e

    return recipes


def load_recipes(
    recipe_dir: Path,
    names_or_tags: Iterable[str],
    load_always_tag: bool = True,
) -> dict[str, MetaConfig]:
    """
    Load the recipes for the given package names or tags.
    Note that this function does not do any dependency resolution.

    Parameters
    ----------
    recipe_dir
        Path to the recipe directory
    names_or_tags
        List of package names or tags to load.
        It also supports the following special values:
            - "*" : all packages
            - "no-numpy-dependents" : all packages except those that depend on numpy (including numpy itself)
    load_always_tag
        Whether to load packages with the "always" tag/

    Returns
    -------
    recipes
        Dictionary of package name => config
    """

    available_recipes = load_all_recipes(recipe_dir)

    # tag => list of recipes with that tag
    tagged_recipes: dict[str, list[MetaConfig]] = {}
    for recipe in available_recipes.values():
        for _tag in recipe.package.tag:
            tagged_recipes.setdefault(_tag, []).append(recipe)

    recipes: dict[str, MetaConfig] = {}

    for name_or_tag in names_or_tags:
        # 1. package name
        if name_or_tag in available_recipes:
            recipes[name_or_tag] = available_recipes[name_or_tag].model_copy(deep=True)

        # 2. tag
        elif (
            name_or_tag.startswith("tag:")
            and (tag := name_or_tag.removeprefix("tag:")) in tagged_recipes
        ):
            for recipe in tagged_recipes[tag]:
                recipes[recipe.package.name] = recipe.model_copy(deep=True)

        # 3. meta packages
        elif name_or_tag == "*":  # all packages
            recipes.update(
                {
                    name: package.model_copy(deep=True)
                    for name, package in available_recipes.items()
                }
            )
        elif name_or_tag == "no-numpy-dependents":
            # This is a meta package and will be handled outside of this function
            recipes["no-numpy-dependents"] = None  # type: ignore[assignment]

        elif name_or_tag in ("core", "min-scipy-stack"):
            logger.warning(
                "Using meta package without the 'tag:' prefix is deprecated, "
                "use 'tag:%s' instead.",
                name_or_tag,
            )
            for recipe in tagged_recipes[name_or_tag]:
                recipes[recipe.package.name] = recipe.model_copy(deep=True)
        else:
            raise ValueError(f"Unknown package name or tag: {name_or_tag}")

    if load_always_tag:
        always_recipes = tagged_recipes.get("always", [])
        for recipe in always_recipes:
            recipes[recipe.package.name] = recipe.model_copy(deep=True)

    return recipes



# This file contains private functions taken from pypa/build.

# Copyright © 2019 Filipe Laíns <filipe.lains@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
import contextlib
import os
import subprocess
import sys
import traceback
import warnings
from collections.abc import Iterator
from typing import NoReturn

from build import (
    BuildBackendException,
    BuildException,
    FailedProcessError,
    ProjectBuilder,
)
from build.env import DefaultIsolatedEnv

_COLORS = {
    "red": "\33[91m",
    "green": "\33[92m",
    "yellow": "\33[93m",
    "bold": "\33[1m",
    "dim": "\33[2m",
    "underline": "\33[4m",
    "reset": "\33[0m",
}
_NO_COLORS = {color: "" for color in _COLORS}


def _init_colors() -> dict[str, str]:
    if "NO_COLOR" in os.environ:
        if "FORCE_COLOR" in os.environ:
            warnings.warn(
                "Both NO_COLOR and FORCE_COLOR environment variables are set, disabling color",
                stacklevel=2,
            )
        return _NO_COLORS
    elif "FORCE_COLOR" in os.environ or sys.stdout.isatty():
        return _COLORS
    return _NO_COLORS


_STYLES = _init_colors()


def _cprint(fmt: str = "", msg: str = "") -> None:
    print(fmt.format(msg, **_STYLES), flush=True)


def _error(msg: str, code: int = 1) -> NoReturn:  # pragma: no cover
    """
    Print an error message and exit. Will color the output when writing to a TTY.

    :param msg: Error message
    :param code: Error code
    """
    _cprint("{red}ERROR{reset} {}", msg)
    raise SystemExit(code)


class _ProjectBuilder(ProjectBuilder):
    @staticmethod
    def log(message: str) -> None:
        _cprint("{bold}* {}{reset}", message)


class _DefaultIsolatedEnv(DefaultIsolatedEnv):
    @staticmethod
    def log(message: str) -> None:
        _cprint("{bold}* {}{reset}", message)

    @property
    def scripts_dir(self) -> str:
        if hasattr(self, "_env_backend"):  # pypabuild >= 1.2.0
            return self._env_backend.scripts_dir
        elif hasattr(self, "_scripts_dir"):
            return self._scripts_dir
        else:
            raise AttributeError("No attribute '_env_backend' or '_scripts_dir' found")


@contextlib.contextmanager
def _handle_build_error() -> Iterator[None]:
    try:
        yield
    except (BuildException, FailedProcessError) as e:
        _error(str(e))
    except BuildBackendException as e:
        if isinstance(e.exception, subprocess.CalledProcessError):
            _cprint()
            _error(str(e))

        if e.exc_info:
            tb_lines = traceback.format_exception(
                e.exc_info[0],
                e.exc_info[1],
                e.exc_info[2],
                limit=-1,
            )
            tb = "".join(tb_lines)
        else:
            tb = traceback.format_exc(-1)  # type: ignore[unreachable]
        _cprint("\n{dim}{}{reset}\n", tb.strip("\n"))
        _error(str(e))



import fnmatch
import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

from packaging.utils import parse_wheel_filename

from pyodide_build.common import (
    chdir,
    modify_wheel,
)


def unvendor_tests_in_wheel(
    wheel: Path, retain_patterns: list[str] | None = None
) -> Path | None:
    """
    Unvendor tests from a wheel file.

    This function finds the tests in the wheel file, and extracts them to a separate
    tar file. The tar file is placed in the same directory as the wheel file, with the -tests.tar suffix.

    Parameters
    ----------
    wheel
        The path to the wheel file.

    retain_patterns
        A list of patterns to retain in the tests. If a pattern is found in the tests, it will be retained.

    Returns
    -------
    The path to the tar file containing the tests. If no tests were found, returns None.
    """
    retain_patterns = retain_patterns or []

    name = parse_wheel_filename(wheel.name)[0]
    file_format = "tar"
    basename = f"{name}-tests"
    fullname = f"{basename}.{file_format}"
    destination = wheel.parent / fullname

    with TemporaryDirectory() as _tmpdir:
        tmpdir = Path(_tmpdir)
        test_dir = tmpdir / "tests"
        with modify_wheel(wheel, verbose=False) as wheel_extract_dir:
            nmoved = unvendor_tests(
                wheel_extract_dir,
                test_dir,
                retain_patterns,
            )
            if not nmoved:
                return None

            with chdir(tmpdir):
                generated_file = shutil.make_archive(basename, file_format, test_dir)
                shutil.move(tmpdir / generated_file, destination)

    return destination


def unvendor_tests(
    install_prefix: Path, test_install_prefix: Path, retain_test_patterns: list[str]
) -> int:
    """Unvendor test files and folders

    This function recursively walks through install_prefix and moves anything
    that looks like a test folder under test_install_prefix.


    Parameters
    ----------
    install_prefix
        the folder where the package was installed
    test_install_prefix
        the folder where to move the tests. If it doesn't exist, it will be
        created.

    Returns
    -------
    n_moved
        number of files or folders moved
    """
    n_moved = 0
    out_files = []
    shutil.rmtree(test_install_prefix, ignore_errors=True)
    for root, _dirs, files in os.walk(install_prefix):
        root_rel = Path(root).relative_to(install_prefix)
        if root_rel.name == "__pycache__" or root_rel.name.endswith(".egg_info"):
            continue
        if root_rel.name in {"test", "tests"}:
            # This is a test folder
            (test_install_prefix / root_rel).parent.mkdir(exist_ok=True, parents=True)
            shutil.move(install_prefix / root_rel, test_install_prefix / root_rel)
            n_moved += 1
            continue
        out_files.append(root)
        for fpath in files:
            if (
                fnmatch.fnmatchcase(fpath, "test_*.py")
                or fnmatch.fnmatchcase(fpath, "*_test.py")
                or fpath == "conftest.py"
            ):
                if any(fnmatch.fnmatchcase(fpath, pat) for pat in retain_test_patterns):
                    continue
                (test_install_prefix / root_rel).mkdir(exist_ok=True, parents=True)
                shutil.move(
                    install_prefix / root_rel / fpath,
                    test_install_prefix / root_rel / fpath,
                )
                n_moved += 1

    return n_moved



import contextlib
import json
import shutil
import subprocess
import tempfile
import urllib.error
import urllib.request
import warnings
from collections.abc import Iterator
from pathlib import Path
from textwrap import indent
from typing import Any, Literal, TypedDict
from urllib import request

from packaging.version import Version
from ruamel.yaml import YAML

from pyodide_build.common import parse_top_level_import_name
from pyodide_build.logger import logger


class URLDict(TypedDict):
    comment_text: str
    digests: dict[str, Any]
    downloads: int
    filename: str
    has_sig: bool
    md5_digest: str
    packagetype: str
    python_version: str
    requires_python: str
    size: int
    upload_time: str
    upload_time_iso_8601: str
    url: str
    yanked: bool
    yanked_reason: str | None


class MetadataDict(TypedDict):
    info: dict[str, Any]
    last_serial: int
    releases: dict[str, list[dict[str, Any]]]
    urls: list[URLDict]
    vulnerabilities: list[Any]


class MkpkgSkipped(Exception):
    pass


class MkpkgFailedException(Exception):
    pass


SDIST_EXTENSIONS = tuple(
    extension
    for (name, extensions, description) in shutil.get_unpack_formats()
    for extension in extensions
)


def _find_sdist(pypi_metadata: MetadataDict) -> URLDict | None:
    """Get sdist file path from the metadata"""
    # The first one we can use. Usually a .tar.gz
    for entry in pypi_metadata["urls"]:
        if entry["packagetype"] == "sdist" and entry["filename"].endswith(
            SDIST_EXTENSIONS
        ):
            return entry
    return None


def _find_wheel(pypi_metadata: MetadataDict, native: bool = False) -> URLDict | None:
    """Get wheel file path from the metadata"""
    predicate = lambda filename: filename.endswith(
        ".whl" if native else "py3-none-any.whl"
    )

    for entry in pypi_metadata["urls"]:
        if entry["packagetype"] == "bdist_wheel" and predicate(entry["filename"]):
            return entry
    return None


def _make_predictable_url(
    package: str, version: str, source_type: Literal["wheel", "sdist"], filename: str
) -> str | None:
    """
    Create a predictable URL for a PyPI package based on PyPI's conventions,
    documented in https://docs.pypi.org/api/#predictable-urls.

    Parameters
    ----------
    package
        The package name
    version
        The package version
    source_type
        Either "wheel" or "sdist"
    filename
        The full filename (used for wheels to extract required tags)

    Returns
    -------
    A predictable URL for the package file, or None if the URL could not be
    constructed.
    """
    from packaging.utils import (
        InvalidWheelFilename,
        canonicalize_name,
        parse_wheel_filename,
    )

    host = "https://files.pythonhosted.org"

    # convert hyphens to underscores for the package name, as
    # PyPI uses underscores in the URL and packaging does not
    # handle this.
    package_name = canonicalize_name(package)
    package_url_name = package_name.replace("-", "_")

    if source_type == "sdist":
        return f"{host}/packages/source/{package_name[0]}/{package_url_name}/{package_url_name}-{version}.tar.gz"

    elif source_type == "wheel":
        try:
            # Special case for universal wheels (py2.py3)
            # we return early as packaging doesn't handle
            # this case properly.
            if "-py2.py3-none-any.whl" in filename:
                python_tag = "py2.py3"
                return f"{host}/packages/{python_tag}/{package_url_name[0]}/{package_url_name}/{filename}"

            _, _, _, tags = parse_wheel_filename(filename)
            python_tag = None
            for tag in tags:
                if (
                    tag.interpreter.startswith("py")
                    and tag.abi == "none"
                    and tag.platform == "any"
                ):
                    python_tag = tag.interpreter
                    break

            if not python_tag:
                msg = f"Not a pure Python wheel: {filename}"
                logger.warning(msg)
                return None

            return f"{host}/packages/{python_tag}/{package_url_name[0]}/{package_url_name}/{filename}"

        except InvalidWheelFilename:
            # Let invalid wheel filenames bubble up
            raise
        except Exception as e:
            msg = f"Error parsing wheel filename {filename}: {e}"
            logger.warning(msg)
            return None

    return None


def _find_dist(
    pypi_metadata: MetadataDict, source_types: list[Literal["wheel", "sdist"]]
) -> URLDict:
    """Find a wheel or sdist, as appropriate.

    source_types controls which types (wheel and/or sdist) are accepted and also
    the priority order.
    E.g., ["wheel", "sdist"] means accept either wheel or sdist but prefer wheel.
    ["sdist", "wheel"] means accept either wheel or sdist but prefer sdist.
    """
    result = None
    for source in source_types:
        if source == "wheel":
            result = _find_wheel(pypi_metadata)
        if source == "sdist":
            result = _find_sdist(pypi_metadata)
        if result:
            package_name = pypi_metadata["info"]["name"]
            version = pypi_metadata["info"]["version"]
            filename = result["filename"]

            predictable_url = _make_predictable_url(
                package_name, version, source, filename
            )
            if predictable_url:
                result_copy = result.copy()
                result_copy["url"] = predictable_url
                return result_copy
            return result

    types_str = " or ".join(source_types)
    name = pypi_metadata["info"].get("name")
    url = pypi_metadata["info"].get("package_url")
    raise MkpkgFailedException(f"No {types_str} found for package {name} ({url})")


def _get_metadata(package: str, version: str | None = None) -> MetadataDict:
    """Download metadata for a package from PyPI"""
    version = ("/" + version) if version is not None else ""
    url = f"https://pypi.org/pypi/{package}{version}/json"

    try:
        with urllib.request.urlopen(url) as fd:
            pypi_metadata = json.load(fd)
    except urllib.error.HTTPError as e:
        raise MkpkgFailedException(
            f"Failed to load metadata for {package}{version} from "
            f"https://pypi.org/pypi/{package}{version}/json: {e}"
        ) from e

    return pypi_metadata


@contextlib.contextmanager
def _download_wheel(pypi_metadata: URLDict) -> Iterator[Path]:
    response = request.urlopen(pypi_metadata["url"])
    whlname = Path(response.geturl()).name

    with tempfile.TemporaryDirectory() as tmpdirname:
        whlpath = Path(tmpdirname, whlname)
        whlpath.write_bytes(response.read())
        yield whlpath


def run_prettier(meta_path: str | Path) -> None:
    try:
        subprocess.run(["npx", "prettier", "-w", meta_path], check=True)
    except FileNotFoundError:
        warnings.warn(
            "'npx' executable missing, output has not been prettified.", stacklevel=1
        )


def load_meta_yaml(yaml, meta_path: Path):
    if not meta_path.exists():
        package = meta_path.parent.name
        logger.error("%s does not exist", meta_path)
        raise MkpkgFailedException(f"{package} recipe not found at {meta_path}")

    return yaml.load(meta_path.read_bytes())


def store_meta_yaml(yaml, meta_path: Path, yaml_content):
    yaml.dump(yaml_content, meta_path)
    run_prettier(meta_path)


def make_package(
    packages_dir: Path,
    package: str,
    version: str | None = None,
    source_fmt: Literal["wheel", "sdist"] | None = None,
    maintainer: str | None = None,
) -> None:
    """
    Creates a template that will work for most pure Python packages,
    but will have to be edited for more complex things.
    """
    logger.info("Creating meta.yaml package for %s", package)

    yaml = YAML()

    pypi_metadata = _get_metadata(package, version)

    if source_fmt:
        sources = [source_fmt]
    else:
        # Prefer wheel unless sdist is specifically requested.
        sources = ["wheel", "sdist"]
    dist_metadata = _find_dist(pypi_metadata, sources)

    native_wheel_metadata = _find_wheel(pypi_metadata, native=True)

    top_level = None
    if native_wheel_metadata is not None:
        with _download_wheel(native_wheel_metadata) as native_wheel_path:
            top_level = parse_top_level_import_name(native_wheel_path)

    url = dist_metadata["url"]
    sha256 = dist_metadata["digests"]["sha256"]
    version = pypi_metadata["info"]["version"]

    homepage = pypi_metadata["info"]["home_page"]
    summary = pypi_metadata["info"]["summary"]
    license = pypi_metadata["info"]["license"]
    pypi = "https://pypi.org/project/" + package

    yaml_content = {
        "package": {
            "name": package,
            "version": version,
            "top-level": top_level or ["PUT_TOP_LEVEL_IMPORT_NAMES_HERE"],
        },
        "source": {"url": url, "sha256": sha256},
        "about": {
            "home": homepage,
            "PyPI": pypi,
            "summary": summary,
            "license": license,
        },
        "extra": {
            "recipe-maintainers": [maintainer or "PUT_YOUR_GITHUB_USERNAME_HERE"]
        },
    }

    package_dir = packages_dir / package
    package_dir.mkdir(parents=True, exist_ok=True)

    meta_path = package_dir / "meta.yaml"
    if meta_path.exists():
        raise MkpkgFailedException(f"The package {package} already exists")

    yaml.representer.ignore_aliases = lambda *_: True
    store_meta_yaml(yaml, meta_path, yaml_content)
    logger.success(f"Output written to {meta_path}")


def update_package(
    root: Path,
    package: str,
    *,
    version: str | None = None,
    update_patched: bool = False,
    update_pinned: bool = False,
    source_fmt: Literal["wheel", "sdist"] | None = None,
) -> None:
    yaml = YAML()

    meta_path = root / package / "meta.yaml"
    yaml_content = load_meta_yaml(yaml, meta_path)

    build_info = yaml_content.get("build", {})
    ty = build_info.get("type", None)
    if ty in ["static_library", "shared_library", "cpython_module"]:
        raise MkpkgSkipped(f"{package} is a {ty.replace('_', ' ')}!")

    if "url" not in yaml_content["source"]:
        raise MkpkgSkipped(f"{package} is a local package!")

    if (not update_pinned) and yaml_content["package"].get("pinned", False):
        raise MkpkgSkipped(f"{package} is pinned!")

    if yaml_content["source"]["url"].endswith("whl"):
        old_fmt = "wheel"
    else:
        old_fmt = "sdist"

    pypi_metadata = _get_metadata(package, version)

    # Grab versions from metadata
    pypi_ver = Version(pypi_metadata["info"]["version"])
    local_ver = Version(yaml_content["package"]["version"])

    # and grab checksums from metadata
    source_fmt = source_fmt or old_fmt
    dist_metadata = _find_dist(pypi_metadata, [source_fmt])
    sha256 = dist_metadata["digests"]["sha256"]
    sha256_local = yaml_content["source"].get("sha256")

    # fail if local version is newer than PyPI version
    # since updating isn't possible in that case
    if pypi_ver < local_ver:
        raise MkpkgFailedException(
            f"Local version {local_ver} is newer than PyPI version {pypi_ver}, "
            f"cannot update {package}. Please verify in case the version was "
            "updated manually and is correct."
        )

    # conditions to check if the package is up to date
    is_sha256_up_to_date = sha256 == sha256_local
    is_version_up_to_date = pypi_ver == local_ver

    already_up_to_date = (is_sha256_up_to_date and is_version_up_to_date) and (
        source_fmt is None or source_fmt == old_fmt
    )
    if already_up_to_date:
        logger.success(
            f"{package} already up to date."
            f" Local: {local_ver} and PyPI: {pypi_ver}"
            f" and checksum received: {sha256} matches local: {sha256_local} ✅"
        )
        return

    logger.info(
        "%s is out of date: either %s < %s or checksums might have mismatched: received %s against local %s 🚨",
        package,
        local_ver,
        pypi_ver,
        sha256,
        sha256_local,
    )

    if yaml_content["source"].get("patches"):
        if update_patched:
            logger.warning(
                "Pyodide applies patches to %s. Update the patches (if needed) to avoid build failing.",
                package,
            )
        else:
            raise MkpkgFailedException(
                f"Pyodide applies patches to {package}. Skipping update."
                f" Use --update-patched to force updating {package}."
            )

    if source_fmt:
        # require the type requested
        sources = [source_fmt]
    elif old_fmt == "wheel":
        # prefer wheel to sdist
        sources = ["wheel", "sdist"]
    else:
        # prefer sdist to wheel
        sources = ["sdist", "wheel"]

    dist_metadata = _find_dist(pypi_metadata, sources)

    yaml_content["source"]["url"] = dist_metadata["url"]
    yaml_content["source"].pop("md5", None)
    yaml_content["source"]["sha256"] = dist_metadata["digests"]["sha256"]
    yaml_content["package"]["version"] = pypi_metadata["info"]["version"]

    store_meta_yaml(yaml, meta_path, yaml_content)

    logger.success(f"Updated {package} from {local_ver} to {pypi_ver}.")


def disable_package(recipe_dir: Path, package: str, message: str) -> None:
    yaml = YAML()

    meta_path = recipe_dir / package / "meta.yaml"
    yaml_content = load_meta_yaml(yaml, meta_path)
    pkg = yaml_content["package"]
    pkg_keys = list(pkg)
    # Insert after the version key
    version_idx = pkg_keys.index("version") + 1
    pkg.insert(version_idx, "_disabled", True)
    # Add message above it
    if message:
        pkg.yaml_set_comment_before_after_key("_disabled", before=message)
    store_meta_yaml(yaml, meta_path, yaml_content)


def remove_comment_on_line(pkg: Any, line: int):
    # Search for comment on the right line. It's probably after the version key
    # where we put it, but this will find it as long as it isn't directly after
    # top_level (we don't traverse the tree to look for it).
    if pkg.ca.comment and pkg.ca.comment.line == line:
        pkg.ca.comment = None
        return
    for cmts in pkg.ca.items.values():
        for idx, cmt in enumerate(cmts):
            if cmt and cmt.start_mark.line == line:
                cmts[idx] = None
                return


def enable_package(recipe_dir: Path, package: str) -> None:
    yaml = YAML()

    meta_path = recipe_dir / package / "meta.yaml"
    yaml_content = load_meta_yaml(yaml, meta_path)

    text_lines = meta_path.read_text().splitlines()
    for idx, line in enumerate(text_lines):  # noqa: B007
        if line.strip().startswith("_disabled"):
            break
    else:
        # Not disabled, let's return
        return

    pkg = yaml_content["package"]
    if text_lines[idx - 1].strip().startswith("#"):
        # There's a comment to remove, we have to hunt it down...
        remove_comment_on_line(pkg, idx - 1)
    del pkg["_disabled"]

    store_meta_yaml(yaml, meta_path, yaml_content)


def pin_package(recipe_dir: Path, package: str, message: str) -> None:
    yaml = YAML()
    meta_path = recipe_dir / package / "meta.yaml"
    # Try to restore the file to its original state. If git isn't installed or
    # the file isn't tracked, just ignore the error.
    subprocess.run(["git", "restore", meta_path], check=False, capture_output=True)
    yaml_content = load_meta_yaml(yaml, meta_path)
    pkg = yaml_content["package"]
    pkg_keys = list(pkg)
    # Insert after the version key
    version_idx = pkg_keys.index("version") + 1
    pkg.insert(version_idx, "pinned", True)
    # Add message above it
    if message:
        pkg.yaml_set_comment_before_after_key("pinned", before=message)
    store_meta_yaml(yaml, meta_path, yaml_content)


def lookup_gh_username():
    try:
        result = subprocess.run(
            ["gh", "auth", "status"], text=True, check=False, capture_output=True
        )
    except FileNotFoundError:
        raise MkpkgFailedException("gh cli is not installed") from None

    def fail(msg):
        if result.stdout:
            msg += "\nstdout:\n"
            msg += indent(result.stdout, "   ")
        if result.stderr:
            msg += "\nstderr:\n"
            msg += indent(result.stderr, "   ")
        raise MkpkgFailedException(msg)

    if result.returncode:
        fail(f"gh auth status failed with status {result.returncode}")

    for line in result.stdout.splitlines():
        if "github.com account" in line:
            break
    else:
        fail("Did not find github.com account in gh auth status")
    return line.partition(" account ")[-1].split()[0]



"""
Builds a Pyodide package.
"""

import fnmatch
import http.client
import os
import shutil
import subprocess
import sys
from collections.abc import Iterator
from datetime import datetime
from email.message import Message
from functools import cache
from pathlib import Path
from typing import Any, cast

import requests
from packaging.utils import parse_wheel_filename

from pyodide_build import common, pypabuild
from pyodide_build.build_env import (
    RUST_BUILD_PRELUDE,
    BuildArgs,
    _create_constraints_file,
    get_build_environment_vars,
    get_build_flag,
    get_pyodide_root,
    get_pyversion_major,
    get_pyversion_minor,
    pyodide_tags,
    replace_so_abi_tags,
    wheel_platform,
)
from pyodide_build.common import (
    _environment_substitute_str,
    _get_sha256_checksum,
    chdir,
    exit_with_stdio,
    find_matching_wheel,
    make_zip_archive,
    modify_wheel,
    retag_wheel,
    retrying_rmtree,
)
from pyodide_build.logger import logger
from pyodide_build.recipe.bash_runner import (
    BashRunnerWithSharedEnvironment,
    get_bash_runner,
)
from pyodide_build.recipe.spec import MetaConfig, _SourceSpec


def _make_whlfile(
    *args: Any, owner: int | None = None, group: int | None = None, **kwargs: Any
) -> str:
    return shutil._make_zipfile(*args, **kwargs)  # type: ignore[attr-defined]


shutil.register_archive_format("whl", _make_whlfile, description="Wheel file")
try:
    shutil.register_unpack_format(
        "whl",
        [".whl", ".wheel"],
        shutil._unpack_zipfile,  # type: ignore[attr-defined]
        description="Wheel file",
    )
except shutil.RegistryError:
    # Error: .whl is already registered for "whl"
    pass


def _extract_tarballname(url: str, headers: dict) -> str:
    tarballname = url.split("/")[-1]

    if "Content-Disposition" in headers:
        msg = Message()
        msg["Content-Disposition"] = headers["Content-Disposition"]

        filename = msg.get_filename()
        if filename is not None:
            tarballname = filename

    return tarballname


def check_versions_match(pkg_name: str, wheel_name: str, version: str):
    wheel_version = str(parse_wheel_filename(wheel_name)[1])
    if wheel_version != version:
        raise ValueError(
            f"Version mismatch in {pkg_name}: version in meta.yaml is '{version}' but version from wheel name is '{wheel_version}'"
        )


class RecipeBuilder:
    """
    A class to build a Pyodide meta.yaml recipe.
    """

    def __init__(
        self,
        recipe: str | Path,
        build_args: BuildArgs,
        build_dir: str | Path | None = None,
        force_rebuild: bool = False,
        continue_: bool = False,
    ):
        """
        Parameters
        ----------
        recipe
            The path to the meta.yaml file or the directory containing
            the meta.yaml file.
        build_args
            The extra build arguments passed to the build script.
        build_dir
            The path to the build directory. By default, it will be
            <the directory containing the meta.yaml file> / build
        force_rebuild
            If True, the package will be rebuilt even if it is already up-to-date.
        continue_
            If True, continue a build from the middle. For debugging. Implies "force_rebuild".
        """
        recipe = Path(recipe).resolve()
        self.pkg_root, self.recipe = _load_recipe(recipe)

        self.name = self.recipe.package.name
        self.version = self.recipe.package.version
        self.fullname = f"{self.name}-{self.version}"

        self.source_metadata = self.recipe.source
        self.build_metadata = self.recipe.build
        self.package_type = self.build_metadata.package_type
        self.is_wheel = self.package_type in ["package", "cpython_module"]

        self.build_dir = (
            Path(build_dir).resolve() if build_dir else self.pkg_root / "build"
        )
        if len(str(self.build_dir).split(maxsplit=1)) > 1:
            raise ValueError(
                "PIP_CONSTRAINT contains spaces so pip will misinterpret it. Make sure the path to the package build directory has no spaces.\n"
                "See https://github.com/pypa/pip/issues/13283"
            )
        self.library_install_prefix = self.build_dir.parent.parent / ".libs"
        self.src_extract_dir = (
            self.build_dir / self.fullname
        )  # where we extract the source

        # where the built artifacts are put.
        # For wheels, this is the default location where the built wheels are put by pypa/build.
        # For shared libraries, users should use this directory to put the built shared libraries (can be accessed by DISTDIR env var)
        self.src_dist_dir = self.src_extract_dir / "dist"

        # where Pyodide will look for the built artifacts when building pyodide-lock.json.
        # after building packages, artifacts in src_dist_dir will be copied to dist_dir
        self.dist_dir = self.pkg_root / "dist"
        self.build_args = build_args
        self.force_rebuild = force_rebuild or continue_
        self.continue_ = continue_

    @classmethod
    def get_builder(
        cls,
        recipe: str | Path,
        build_args: BuildArgs,
        build_dir: str | Path | None = None,
        force_rebuild: bool = False,
        continue_: bool = False,
    ) -> "RecipeBuilder":
        recipe = Path(recipe).resolve()
        _, config = _load_recipe(recipe)
        match config.build.package_type:
            case "package" | "cpython_module":
                builder = RecipeBuilderPackage
            case "static_library":
                builder = RecipeBuilderStaticLibrary
            case "shared_library":
                builder = RecipeBuilderSharedLibrary
            case _:
                raise ValueError(f"Unknown package type: {config.build.package_type}")

        return builder(recipe, build_args, build_dir, force_rebuild, continue_)

    def build(self) -> None:
        """
        Build the package. This is the only public method of this class.
        """
        self._check_executables()

        t0 = datetime.now()
        timestamp = t0.strftime("%Y-%m-%d %H:%M:%S")
        logger.info("[%s] Building package %s...", timestamp, self.name)
        success = True
        try:
            self._build()

            if not self.is_wheel:
                (self.build_dir / ".packaged").touch()
        except (Exception, KeyboardInterrupt):
            success = False
            raise
        except SystemExit as e:
            success = e.code == 0
            raise
        finally:
            t1 = datetime.now()
            datestamp = "[{}]".format(t1.strftime("%Y-%m-%d %H:%M:%S"))
            total_seconds = f"{(t1 - t0).total_seconds():.1f}"
            status = "Succeeded" if success else "Failed"
            msg = f"{datestamp} {status} building package {self.name} in {total_seconds} seconds."
            if success:
                logger.success(msg)
            else:
                logger.error(msg)

    def _build_package(self, bash_runner: BashRunnerWithSharedEnvironment) -> None:
        raise NotImplementedError("Subclasses must implement this method")

    def _build(self) -> None:
        if not self.force_rebuild and not needs_rebuild(
            self.pkg_root,
            self.build_dir,
            self.source_metadata,
            self.is_wheel,
            self.version,
        ):
            return

        if self.continue_ and not self.src_extract_dir.exists():
            raise OSError(
                "Cannot find source for rebuild. Expected to find the source "
                f"directory at the path {self.src_extract_dir}, but that path does not exist."
            )

        self._redirect_stdout_stderr_to_logfile()

        if not self.continue_:
            self._prepare_source()
            self._patch()

        with (
            chdir(self.pkg_root),
            get_bash_runner(self._get_helper_vars() | os.environ.copy()) as bash_runner,
        ):
            self._build_package(bash_runner)

    def _check_executables(self) -> None:
        """
        Check that the executables required to build the package are available.
        """
        missing_executables = common.find_missing_executables(
            self.recipe.requirements.executable
        )
        if missing_executables:
            missing_string = ", ".join(missing_executables)
            error_msg = (
                f"The following executables are required to build {self.name}, but missing in the host system: "
                + missing_string
            )
            raise RuntimeError(error_msg)

    def _prepare_source(self) -> None:
        """
        Figure out from the "source" key in the package metadata where to get the source
        from, then get the source into the build directory.
        """

        # clear the build directory
        if self.build_dir.resolve().is_dir():
            retrying_rmtree(self.build_dir)

        self.build_dir.mkdir(parents=True, exist_ok=True)

        if self.source_metadata.url is not None:
            self._download_and_extract()
            return

        # Build from local source, mostly for testing purposes.
        if self.source_metadata.path is None:
            raise ValueError(
                "Incorrect source provided. Either a url or a path must be provided."
            )

        srcdir = self.source_metadata.path.resolve()
        if not srcdir.is_dir():
            raise ValueError(f"path={srcdir} must point to a directory that exists")

        def ignore(path: str, names: list[str]) -> list[str]:
            ignored: list[str] = []

            if fnmatch.fnmatch(path, "*/dist"):
                # Do not copy dist/*.whl files from a dirty source tree;
                # this can lead to "Exception: Unexpected number of wheels" later.
                ignored.extend(name for name in names if name.endswith(".whl"))
            return ignored

        shutil.copytree(srcdir, self.src_extract_dir, ignore=ignore)

        self.src_dist_dir.mkdir(parents=True, exist_ok=True)

    def _download_and_extract(self) -> None:
        """
        Download the source from specified in the package metadata,
        then checksum it, then extract the archive into the build directory.
        """
        build_env = get_build_environment_vars(get_pyodide_root())
        url = cast(str, self.source_metadata.url)  # we know it's not None
        url = _environment_substitute_str(url, build_env)

        max_retry = 3
        for retry_cnt in range(max_retry):
            try:
                response = requests.get(url)
                response.raise_for_status()
            except (
                requests.exceptions.RequestException,
                http.client.HTTPException,
            ) as e:
                if retry_cnt == max_retry - 1:
                    raise RuntimeError(
                        f"Failed to download {url} after {max_retry} trials"
                    ) from e

                continue

            break

        self.build_dir.mkdir(parents=True, exist_ok=True)

        tarballname = _extract_tarballname(url, response.headers)
        tarballpath = self.build_dir / tarballname
        tarballpath.write_bytes(response.content)

        checksum = self.source_metadata.sha256
        if checksum is not None:
            try:
                checksum = _environment_substitute_str(checksum, build_env)
                check_checksum(tarballpath, checksum)
            except Exception:
                tarballpath.unlink()
                raise

        # already built
        if tarballpath.suffix == ".whl":
            check_versions_match(self.name, tarballpath.name, self.version)
            self.src_dist_dir.mkdir(parents=True, exist_ok=True)
            shutil.copy(tarballpath, self.src_dist_dir)
            return

        # Use a Python 3.14-like filter (see https://github.com/python/cpython/issues/112760)
        # Can be removed once we use Python 3.14
        # The "data" filter will reset ownership but preserve permissions and modification times
        # Without it, permissions and modification times will be silently skipped if the uid/git
        # is too large for the chown() call. This behavior can lead to "Permission denied" errors
        # (missing x bit) or random strange `make` behavior (due to wrong mtime order) in the CI
        # pipeline.
        shutil.unpack_archive(
            tarballpath,
            self.build_dir,
            filter=None if tarballpath.suffix == ".zip" else "data",
        )

        extract_dir_name = self.source_metadata.extract_dir
        if extract_dir_name is None:
            extract_dir_name = trim_archive_extension(tarballname)

        shutil.move(self.build_dir / extract_dir_name, self.src_extract_dir)
        self.src_dist_dir.mkdir(parents=True, exist_ok=True)

    def _create_constraints_file(self) -> str:
        """
        Creates a pip constraints file by concatenating global constraints (PIP_CONSTRAINT)
        with constraints specific to this package.

        returns the path to the new constraints file.
        """
        host_constraints = _create_constraints_file()

        constraints = self.recipe.requirements.constraint
        if not constraints:
            # nothing to override
            return host_constraints

        new_constraints_file = self.build_dir / "constraints.txt"
        with new_constraints_file.open("w") as f:
            for constraint in constraints:
                f.write(constraint + "\n")

        return host_constraints + " " + str(new_constraints_file)

    def _compile(
        self,
        bash_runner: BashRunnerWithSharedEnvironment,
    ) -> None:
        """
        Runs pypa/build for the package.

        Parameters
        ----------
        bash_runner
            The runner we will use to execute our bash commands. Preserves environment
            variables from one invocation to the next.

        target_install_dir
            The path to the target Python installation

        """

        cflags = self.build_metadata.cflags + " " + self.build_args.cflags
        cxxflags = self.build_metadata.cxxflags + " " + self.build_args.cxxflags
        ldflags = self.build_metadata.ldflags + " " + self.build_args.ldflags

        build_env_ctx = pypabuild.get_build_env(
            env=bash_runner.env,
            pkgname=self.name,
            cflags=cflags,
            cxxflags=cxxflags,
            ldflags=ldflags,
            target_install_dir=self.build_args.target_install_dir,
            exports=self.build_metadata.exports,
            build_dir=self.build_dir,
        )
        config_settings = pypabuild.parse_backend_flags(
            self.build_metadata.backend_flags
        )

        with build_env_ctx as build_env:
            if self.build_metadata.cross_script is not None:
                with BashRunnerWithSharedEnvironment(build_env) as runner:
                    runner.run(
                        self.build_metadata.cross_script,
                        script_name="cross script",
                        cwd=self.src_extract_dir,
                    )
                    build_env = runner.env

            build_env["PIP_CONSTRAINT"] = str(self._create_constraints_file())

            wheel_path = pypabuild.build(
                self.src_extract_dir, self.src_dist_dir, build_env, config_settings
            )
            check_versions_match(self.name, Path(wheel_path).name, self.version)

    def _patch(self) -> None:
        """
        Apply patches to the source.
        """
        token_path = self.src_extract_dir / ".patched"
        if token_path.is_file():
            return

        patches = self.source_metadata.patches
        extras = self.source_metadata.extras
        cast(str, self.source_metadata.url)

        if not patches and not extras:
            return

        # Apply all the patches
        for patch in patches:
            patch_abspath = self.pkg_root / patch
            result = subprocess.run(
                ["patch", "-p1", "--binary", "--verbose", "-i", patch_abspath],
                check=False,
                encoding="utf-8",
                cwd=self.src_extract_dir,
            )
            if result.returncode != 0:
                logger.error("ERROR: Patch %s failed", patch_abspath)
                exit_with_stdio(result)

        # Add any extra files
        for src, dst in extras:
            shutil.copyfile(self.pkg_root / src, self.src_extract_dir / dst)

        token_path.touch()

    def _redirect_stdout_stderr_to_logfile(self) -> None:
        """
        Redirect stdout and stderr to a log file.
        """
        try:
            stdout_fileno = sys.stdout.fileno()
            stderr_fileno = sys.stderr.fileno()

            tee = subprocess.Popen(
                ["tee", self.pkg_root / "build.log"], stdin=subprocess.PIPE
            )

            # Cause tee's stdin to get a copy of our stdin/stdout (as well as that
            # of any child processes we spawn)
            os.dup2(tee.stdin.fileno(), stdout_fileno)  # type: ignore[union-attr]
            os.dup2(tee.stdin.fileno(), stderr_fileno)  # type: ignore[union-attr]
        except OSError:
            # This normally happens when testing
            logger.warning("stdout/stderr does not have a fileno, not logging to file")

    def _get_helper_vars(self) -> dict[str, str]:
        """
        Get the helper variables for the build script.
        """
        return {
            "PKGDIR": str(self.pkg_root),
            "PKG_VERSION": self.version,
            "PKG_BUILD_DIR": str(self.src_extract_dir),
            "DISTDIR": str(self.src_dist_dir),
            # TODO: rename this to something more compatible with Makefile or CMake conventions
            "WASM_LIBRARY_DIR": str(self.library_install_prefix),
            # Emscripten will use this variable to configure pkg-config in emconfigure
            "EM_PKG_CONFIG_PATH": str(self.library_install_prefix / "lib/pkgconfig"),
            # This variable is usually overwritten by emconfigure
            # The value below will only be used if pkg-config is called without emconfigure
            # We use PKG_CONFIG_LIBDIR instead of PKG_CONFIG_PATH,
            # so pkg-config will not look in the default system directories
            "PKG_CONFIG_LIBDIR": str(self.library_install_prefix / "lib/pkgconfig"),
        }


class RecipeBuilderPackage(RecipeBuilder):
    """
    Recipe builder for python packages.
    """

    def _build_package(self, bash_runner: BashRunnerWithSharedEnvironment) -> None:
        if self.recipe.is_rust_package():
            bash_runner.run(
                RUST_BUILD_PRELUDE,
                script_name="rust build prelude",
                cwd=self.src_extract_dir,
            )

        bash_runner.run(
            self.build_metadata.script,
            script_name="build script",
            cwd=self.src_extract_dir,
        )

        url = self.source_metadata.url
        prebuilt_wheel = url and url.endswith(".whl")
        if not prebuilt_wheel:
            self._compile(bash_runner)

        self._package_wheel(bash_runner)
        shutil.copytree(self.src_dist_dir, self.dist_dir, dirs_exist_ok=True)

    def _package_wheel(
        self,
        bash_runner: BashRunnerWithSharedEnvironment,
    ) -> None:
        """Package a wheel

        This unpacks the wheel, runs and "build.post"
        script, and then repacks the wheel.

        Parameters
        ----------
        bash_runner
            The runner we will use to execute our bash commands. Preserves
            environment variables from one invocation to the next.
        """
        wheel = find_matching_wheel(
            self.src_dist_dir.glob("*.whl"), pyodide_tags(), version=self.version
        )
        if not wheel:
            raise RuntimeError(
                f"Found no wheel while building {self.name}. Candidates:\n"
                + "\n".join(f.name for f in self.src_dist_dir.glob("*.whl"))
            )

        if self.package_type == "cpython_module":
            abi = f"cp{get_pyversion_major()}{get_pyversion_minor()}"
            wheel = retag_wheel(wheel, wheel_platform(), python=abi, abi=abi)
        elif "emscripten" in wheel.name:
            # Retag platformed wheels to pyodide
            wheel = retag_wheel(wheel, wheel_platform())

        logger.info("Unpacking wheel to %s", str(wheel))

        name, ver, _ = wheel.name.split("-", 2)

        with modify_wheel(wheel) as wheel_dir:
            # update so abi tags after build is complete but before running post script
            # to maximize sanity.
            replace_so_abi_tags(wheel_dir)
            bash_runner.run(
                self.build_metadata.post, script_name="post script", cwd=wheel_dir
            )

            if self.build_metadata.vendor_sharedlib:
                lib_dir = self.library_install_prefix
                # Old version of Emscripten does not have RUNTIME_PATH section, so only
                # patch the rpath if it is present.
                should_modify_rpath = get_build_flag("PYODIDE_ABI_VERSION") > "2025"
                copy_sharedlibs(
                    wheel, wheel_dir, lib_dir, modify_rpath=should_modify_rpath
                )

            python_dir = f"python{sys.version_info.major}.{sys.version_info.minor}"
            host_site_packages = (
                Path(self.build_args.host_install_dir)
                / f"lib/{python_dir}/site-packages"
            )
            if self.build_metadata.cross_build_env:
                subprocess.run(
                    [
                        "pip",
                        "install",
                        # Upgrade the package in the host environment if there is a
                        # older or newer version of the package already installed.
                        # However, we don't want to replace the dependencies of the package,
                        # since they may contain cross-build files as well.
                        # For instance, numpy and scipy are both cross-build packages, but scipy depends on numpy.
                        # Therefore, if we install numpy first and then scipy, installing scipy
                        # will overwrite the cross build files in numpy.
                        "--upgrade",
                        "--no-deps",
                        "-t",
                        str(host_site_packages),
                        f"{name}=={ver}",
                    ],
                    check=True,
                )

                # Call the same pip command again to install the dependencies
                # but without the --upgrade flag. This will prevent pip from
                # overwriting the dependencies in the host environment.
                subprocess.run(
                    [
                        "pip",
                        "install",
                        "-t",
                        str(host_site_packages),
                        f"{name}=={ver}",
                    ],
                    check=True,
                )

            for cross_build_file in self.build_metadata.cross_build_files:
                shutil.copy(
                    (wheel_dir / cross_build_file),
                    host_site_packages / cross_build_file,
                )


class RecipeBuilderStaticLibrary(RecipeBuilder):
    """
    Recipe builder for static libraries.
    """

    def _build_package(self, bash_runner: BashRunnerWithSharedEnvironment) -> None:
        bash_runner.run(
            self.build_metadata.script,
            script_name="build script",
            cwd=self.src_extract_dir,
        )


class RecipeBuilderSharedLibrary(RecipeBuilder):
    """
    Recipe builder for shared libraries.
    """

    def _build_package(self, bash_runner: BashRunnerWithSharedEnvironment) -> None:
        bash_runner.run(
            self.build_metadata.script,
            script_name="build script",
            cwd=self.src_extract_dir,
        )

        # copy .so files to dist_dir
        # and create a zip archive of the .so files
        shutil.rmtree(self.dist_dir, ignore_errors=True)
        self.dist_dir.mkdir(parents=True)
        make_zip_archive(self.dist_dir / f"{self.fullname}.zip", self.src_dist_dir)


@cache
def _load_recipe(package_dir: Path) -> tuple[Path, MetaConfig]:
    """
    Load the package configuration from the given directory.

    Parameters
    ----------
    package_dir
        The directory containing the package configuration, or the path to the
        package configuration file.

    Returns
    -------
    pkg_dir
        The directory containing the package configuration.
    pkg
        The package configuration.
    """
    if not package_dir.exists():
        raise FileNotFoundError(f"Package directory {package_dir} does not exist")

    if package_dir.is_dir():
        meta_file = package_dir / "meta.yaml"
    else:
        meta_file = package_dir
        package_dir = meta_file.parent

    return package_dir, MetaConfig.from_yaml(meta_file)


def check_checksum(archive: Path, checksum: str) -> None:
    """
    Checks that an archive matches the checksum in the package metadata.


    Parameters
    ----------
    archive
        the path to the archive we wish to checksum
    checksum
        the checksum we expect the archive to have
    """
    real_checksum = _get_sha256_checksum(archive)
    if real_checksum != checksum:
        raise ValueError(
            f"Invalid sha256 checksum: {real_checksum} != {checksum} (expected)"
        )


def trim_archive_extension(tarballname: str) -> str:
    for extension in [
        ".tar.gz",
        ".tgz",
        ".tar",
        ".tar.bz2",
        ".tbz2",
        ".tar.xz",
        ".txz",
        ".zip",
        ".whl",
    ]:
        if tarballname.endswith(extension):
            return tarballname[: -len(extension)]
    return tarballname


def copy_sharedlibs(
    wheel_file: Path,
    wheel_dir: Path,
    lib_dir: Path,
    modify_rpath=False,
) -> dict[str, Path]:
    from auditwheel_emscripten import copylib, modify_runtime_path, resolve_sharedlib
    from auditwheel_emscripten.wheel_utils import WHEEL_INFO_RE

    match = WHEEL_INFO_RE.match(wheel_file.name)
    if match is None:
        raise RuntimeError(f"Failed to parse wheel file name: {wheel_file.name}")

    dep_map: dict[str, Path] = resolve_sharedlib(
        wheel_dir,
        lib_dir,
    )
    lib_sdir: str = match.group("name") + ".libs"
    if dep_map:
        dep_map_new = copylib(wheel_dir, dep_map, lib_sdir)
        if modify_rpath:
            modify_runtime_path(wheel_dir, lib_sdir)
        logger.info("Copied shared libraries:")
        for lib, path in dep_map_new.items():
            original_path = dep_map[lib]
            logger.info("  %s -> %s", original_path, path)

        return dep_map_new

    return {}


# TODO: move this to common.py or somewhere else
def needs_rebuild(
    pkg_root: Path,
    buildpath: Path,
    source_metadata: _SourceSpec,
    is_wheel: bool = True,
    version: str | None = None,
) -> bool:
    """
    Determines if a package needs a rebuild because its meta.yaml, patches, or
    sources are newer than the `.packaged` thunk.

    pkg_root
        The path to the root directory for the package. Generally
        $PYODIDE_ROOT/packages/<PACKAGES>

    buildpath
        The path to the build directory. By default, it will be
        $(PYOIDE_ROOT)/packages/<PACKAGE>/build/.

    src_metadata
        The source section from meta.yaml.
    """
    dist_dir = pkg_root / "dist"
    if is_wheel:
        previous_wheel = find_matching_wheel(
            dist_dir.glob("*.whl"), pyodide_tags(), version=version
        )
        if not previous_wheel:
            return True
        package_time = previous_wheel.stat().st_mtime
    else:
        packaged_token = buildpath / ".packaged"
        if not packaged_token.is_file():
            logger.debug(
                "%s needs rebuild because %s does not exist", pkg_root, packaged_token
            )
            return True
        package_time = packaged_token.stat().st_mtime

    def source_files() -> Iterator[Path]:
        yield pkg_root / "meta.yaml"
        yield from (pkg_root / patch_path for patch_path in source_metadata.patches)
        yield from (pkg_root / patch_path for [patch_path, _] in source_metadata.extras)
        src_path = source_metadata.path
        if src_path:
            yield from (pkg_root / src_path).resolve().glob("**/*")

    for source_file in source_files():
        source_file = Path(source_file)
        if source_file.stat().st_mtime > package_time:
            return True
    return False



import os
import subprocess
from collections.abc import Mapping
from copy import deepcopy
from pathlib import Path
from types import MappingProxyType

from pyodide_build.common import (
    _environment_substitute_str,
    exit_with_stdio,
    search_pyproject_toml,
)
from pyodide_build.constants import BASE_IGNORED_REQUIREMENTS
from pyodide_build.logger import logger


class ConfigManager:
    """
    Configuration manager for pyodide-build.
    This class works "before" installing the cross build environment.
    So it does not have access to the variables that are retrieved from the cross build environment.

    Most of the times, use CrossBuildEnvConfigManager instead of this class.
    But if you need to access the configuration without installing the cross build environment, use this class.
    """

    def __init__(self):
        self._config = {
            **self._load_default_config(),
            **self._load_cross_build_envs(),
            **self._load_config_file(Path.cwd(), os.environ),
            **self._load_config_from_env(os.environ),
        }

    def _load_default_config(self) -> Mapping[str, str]:
        return deepcopy(DEFAULT_CONFIG)

    def _load_cross_build_envs(self) -> Mapping[str, str]:
        """
        Load environment variables from the cross build environment.
        """

        # This method should be implemented in the subclass.
        return {}

    def _load_config_from_env(self, env: Mapping[str, str]) -> Mapping[str, str]:
        return {
            BUILD_VAR_TO_KEY[key]: env[key] for key in env if key in BUILD_VAR_TO_KEY
        }

    def _load_config_file(
        self, curdir: Path, env: Mapping[str, str]
    ) -> Mapping[str, str]:
        pyproject_path, configs = search_pyproject_toml(curdir)

        if pyproject_path is None or configs is None:
            return {}

        if (
            "tool" in configs
            and "pyodide" in configs["tool"]
            and "build" in configs["tool"]["pyodide"]
        ):
            build_config = {}
            for key, v in configs["tool"]["pyodide"]["build"].items():
                if key not in OVERRIDABLE_BUILD_KEYS:
                    logger.warning(
                        "WARNING: The provided build key %s is either invalid or not overridable, hence ignored.",
                        key,
                    )
                    continue
                build_config[key] = _environment_substitute_str(v, env)

            return build_config
        else:
            return {}

    @property
    def config(self) -> Mapping[str, str]:
        return MappingProxyType(self._config)

    def to_env(self) -> dict[str, str]:
        """
        Export the configuration to environment variables.
        """
        return {BUILD_KEY_TO_VAR[k]: v for k, v in self.config.items()}


class CrossBuildEnvConfigManager(ConfigManager):
    """
    Configuration manager for Package build process.
    This class works "after" installing the cross build environment.

    The configuration manager is responsible for loading configuration from various sources.
    The configuration can be loaded from the following sources (in order of precedence):

        1. Command line arguments (TODO)
        2. Environment variables
        3. Configuration file
        4. Makefile.envs
        5. Default values
    """

    def __init__(self, pyodide_root: Path):
        self.pyodide_root = pyodide_root
        super().__init__()

    def _load_cross_build_envs(self) -> Mapping[str, str]:
        makefile_vars = self._get_make_environment_vars()
        computed_vars = {
            k: _environment_substitute_str(v, env=makefile_vars)
            for k, v in DEFAULT_CONFIG_COMPUTED.items()
        }

        return {
            BUILD_VAR_TO_KEY[k]: v
            for k, v in makefile_vars.items()
            if k in BUILD_VAR_TO_KEY
        } | computed_vars

    def _get_make_environment_vars(self) -> Mapping[str, str]:
        """
        Load environment variables from Makefile.envs
        """
        environment = {}
        result = subprocess.run(
            ["make", "-f", str(self.pyodide_root / "Makefile.envs"), ".output_vars"],
            capture_output=True,
            text=True,
            env={"PYODIDE_ROOT": str(self.pyodide_root)},
            check=False,
        )

        if result.returncode != 0:
            logger.error(
                "ERROR: Failed to load environment variables from Makefile.envs"
            )
            exit_with_stdio(result)

        for line in result.stdout.splitlines():
            equalPos = line.find("=")
            if equalPos != -1:
                varname = line[0:equalPos]

                if varname not in BUILD_VAR_TO_KEY:
                    continue

                value = line[equalPos + 1 :]
                value = value.strip("'").strip()
                environment[varname] = value

        return environment


# Configuration variables and corresponding environment variables.
# TODO: distinguish between variables that are overridable by the user and those that are not.
BUILD_KEY_TO_VAR: dict[str, str] = {
    "pyodide_version": "PYODIDE_VERSION",
    "pyodide_abi_version": "PYODIDE_ABI_VERSION",
    "cargo_build_target": "CARGO_BUILD_TARGET",
    "cargo_target_wasm32_unknown_emscripten_linker": "CARGO_TARGET_WASM32_UNKNOWN_EMSCRIPTEN_LINKER",
    "host_install_dir": "HOSTINSTALLDIR",
    "host_site_packages": "HOSTSITEPACKAGES",
    "numpy_lib": "NUMPY_LIB",
    "pyodide_interpreter": "PYODIDE_INTERPRETER",
    "pyodide_package_index": "PYODIDE_PACKAGE_INDEX",
    "platform_triplet": "PLATFORM_TRIPLET",
    "pip_constraint": "PIP_CONSTRAINT",
    "pymajor": "PYMAJOR",
    "pymicro": "PYMICRO",
    "pyminor": "PYMINOR",
    "pyo3_cross_include_dir": "PYO3_CROSS_INCLUDE_DIR",
    "pyo3_cross_lib_dir": "PYO3_CROSS_LIB_DIR",
    "pyo3_cross_python_version": "PYO3_CROSS_PYTHON_VERSION",
    "pyodide_emscripten_version": "PYODIDE_EMSCRIPTEN_VERSION",
    "pyodide_jobs": "PYODIDE_JOBS",
    "pyodide_root": "PYODIDE_ROOT",
    "python_archive_sha256": "PYTHON_ARCHIVE_SHA256",
    "python_archive_url": "PYTHON_ARCHIVE_URL",
    "pythoninclude": "PYTHONINCLUDE",
    "pyversion": "PYVERSION",
    "cpythoninstall": "CPYTHONINSTALL",
    "rustflags": "RUSTFLAGS",
    "rust_toolchain": "RUST_TOOLCHAIN",
    "rust_emscripten_target_url": "RUST_EMSCRIPTEN_TARGET_URL",
    "cflags": "SIDE_MODULE_CFLAGS",
    "cxxflags": "SIDE_MODULE_CXXFLAGS",
    "ldflags": "SIDE_MODULE_LDFLAGS",
    "stdlib_module_cflags": "STDLIB_MODULE_CFLAGS",
    "sysconfigdata_dir": "SYSCONFIGDATA_DIR",
    "sysconfig_name": "SYSCONFIG_NAME",
    "targetinstalldir": "TARGETINSTALLDIR",
    "cmake_toolchain_file": "CMAKE_TOOLCHAIN_FILE",
    "meson_cross_file": "MESON_CROSS_FILE",
    "cflags_base": "CFLAGS_BASE",
    "cxxflags_base": "CXXFLAGS_BASE",
    "ldflags_base": "LDFLAGS_BASE",
    "home": "HOME",
    "path": "PATH",
    "zip_compression_level": "PYODIDE_ZIP_COMPRESSION_LEVEL",
    "skip_emscripten_version_check": "SKIP_EMSCRIPTEN_VERSION_CHECK",
    "build_dependency_index_url": "BUILD_DEPENDENCY_INDEX_URL",
    "default_cross_build_env_url": "DEFAULT_CROSS_BUILD_ENV_URL",
    "xbuildenv_path": "PYODIDE_XBUILDENV_PATH",
    "ignored_build_requirements": "IGNORED_BUILD_REQUIREMENTS",
    # maintainer only
    "_f2c_fixes_wrapper": "_F2C_FIXES_WRAPPER",
}

BUILD_VAR_TO_KEY = {v: k for k, v in BUILD_KEY_TO_VAR.items()}

# Configuration keys that can be overridden by the user.
# TODO: distinguish between variables that are overridable by the user and those that are not.
OVERRIDABLE_BUILD_KEYS = {
    "cflags",
    "cxxflags",
    "ldflags",
    "rust_toolchain",
    "rust_emscripten_target_url",
    "meson_cross_file",
    "skip_emscripten_version_check",
    "build_dependency_index_url",
    "default_cross_build_env_url",
    "xbuildenv_path",
    "ignored_build_requirements",
    # maintainer only
    "_f2c_fixes_wrapper",
}

# Default configuration values.
TOOLS_DIR = Path(__file__).parent / "tools"
DEFAULT_CONFIG: dict[str, str] = {
    # Paths to toolchain configuration files
    "cmake_toolchain_file": str(TOOLS_DIR / "cmake/Modules/Platform/Emscripten.cmake"),
    "meson_cross_file": str(TOOLS_DIR / "emscripten.meson.cross"),
    # Rust-specific configuration
    "rustflags": "-C link-arg=-sSIDE_MODULE=2 -C link-arg=-sWASM_BIGINT -Z link-native-libraries=no",
    "cargo_build_target": "wasm32-unknown-emscripten",
    "cargo_target_wasm32_unknown_emscripten_linker": "emcc",
    "rust_toolchain": "nightly-2025-02-01",
    "rust_emscripten_target_url": "",
    # Other configuration
    "pyodide_jobs": "1",
    "skip_emscripten_version_check": "0",
    "build_dependency_index_url": "https://pypi.anaconda.org/pyodide/simple",
    "default_cross_build_env_url": "",
    "xbuildenv_path": "",
    # A list of PEP508 build-time requirements to be ignored when building a wheel
    "ignored_build_requirements": " ".join(BASE_IGNORED_REQUIREMENTS),
    # maintainer only
    "_f2c_fixes_wrapper": "",
}

# Default configs that are computed from other values (often from Makefile.envs)
# TODO: Remove dependency on Makefile.envs
DEFAULT_CONFIG_COMPUTED: dict[str, str] = {
    # Compiler flags
    "cflags": "$(CFLAGS_BASE) -I$(PYTHONINCLUDE)",
    "cxxflags": "$(CXXFLAGS_BASE)",
    "ldflags": "$(LDFLAGS_BASE) -s SIDE_MODULE=1",
    # Rust-specific configuration
    "pyo3_cross_lib_dir": "$(CPYTHONINSTALL)/sysconfigdata",  # FIXME: pyodide xbuildenv stores sysconfigdata here
    "pyo3_cross_include_dir": "$(PYTHONINCLUDE)",
    "pyo3_cross_python_version": "$(PYMAJOR).$(PYMINOR)",
    # Misc
    "stdlib_module_cflags": "$(CFLAGS_BASE) -I$(PYTHONINCLUDE) -I Include/ -I. -IInclude/internal/",  # TODO: remove this
    # Paths to build dependencies
    "host_install_dir": "$(PYODIDE_ROOT)/packages/.artifacts",
    "host_site_packages": "$(PYODIDE_ROOT)/packages/.artifacts/lib/python$(PYMAJOR).$(PYMINOR)/site-packages",
    "numpy_lib": "$(PYODIDE_ROOT)/packages/.artifacts/lib/python$(PYMAJOR).$(PYMINOR)/site-packages/numpy/",
    "pyodide_interpreter": "$(PYODIDE_ROOT)/dist/python",
    "pyodide_package_index": "$(PYODIDE_ROOT)/package_index",
}


# A dictionary of config variables that are exposed through pyodide config CLI.
PYODIDE_CLI_CONFIGS = {
    "emscripten_version": "PYODIDE_EMSCRIPTEN_VERSION",
    "python_version": "PYVERSION",
    "rustflags": "RUSTFLAGS",
    "cmake_toolchain_file": "CMAKE_TOOLCHAIN_FILE",
    "rust_toolchain": "RUST_TOOLCHAIN",
    "rust_emscripten_target_url": "RUST_EMSCRIPTEN_TARGET_URL",
    "cflags": "SIDE_MODULE_CFLAGS",
    "cxxflags": "SIDE_MODULE_CXXFLAGS",
    "ldflags": "SIDE_MODULE_LDFLAGS",
    "meson_cross_file": "MESON_CROSS_FILE",
    "xbuildenv_path": "PYODIDE_XBUILDENV_PATH",
    "pyodide_abi_version": "PYODIDE_ABI_VERSION",
    "pyodide_root": "PYODIDE_ROOT",
    "python_include_dir": "PYTHONINCLUDE",
    "ignored_build_requirements": "IGNORED_BUILD_REQUIREMENTS",
    "interpreter": "PYODIDE_INTERPRETER",
    "package_index": "PYODIDE_PACKAGE_INDEX",
}



# This file contains functions for managing the Pyodide build environment.

import dataclasses
import functools
import os
import re
import subprocess
import sys
from collections.abc import Iterator, Sequence
from contextlib import nullcontext, redirect_stdout
from io import StringIO
from pathlib import Path

from packaging.tags import Tag, compatible_tags, cpython_tags

from pyodide_build import __version__
from pyodide_build.common import default_xbuildenv_path, search_pyproject_toml, to_bool

RUST_BUILD_PRELUDE = """
rustup default ${RUST_TOOLCHAIN}
"""


@dataclasses.dataclass(eq=False, order=False, kw_only=True)
class BuildArgs:
    """
    Common arguments for building a package.
    """

    pkgname: str = ""
    cflags: str = ""
    cxxflags: str = ""
    ldflags: str = ""
    target_install_dir: str = ""  # The path to the target Python installation
    host_install_dir: str = ""  # Directory for installing built host packages.
    builddir: str = ""  # The path to run pypa/build


def init_environment(
    *, quiet: bool = False, xbuildenv_path: Path | None = None
) -> None:
    """
    Initialize Pyodide build environment.
    This function needs to be called before any other Pyodide build functions.

    Parameters
    ----------
    quiet
        If True, do not print any messages
    xbuildenv_path
        Path to the cross-build environment directory. If None, the default
        location will be used.
    """

    # Already initialized
    if "PYODIDE_ROOT" in os.environ:
        return

    root = search_pyodide_root(Path.cwd())
    if not root:  # Not in Pyodide tree
        root = _init_xbuild_env(quiet=quiet, xbuildenv_path=xbuildenv_path)

    os.environ["PYODIDE_ROOT"] = str(root)


def _init_xbuild_env(
    *, quiet: bool = False, xbuildenv_path: Path | None = None
) -> Path:
    """
    Initialize the build environment for out-of-tree builds.

    Parameters
    ----------
    quiet
        If True, do not print any messages
    xbuildenv_path
        Path to the cross-build environment directory. If None, the default location will be used.

    Returns
    -------
        The path to the Pyodide root directory inside the xbuild environment
    """
    from pyodide_build.xbuildenv import CrossBuildEnvManager  # avoid circular import

    xbuildenv_path = xbuildenv_path or default_xbuildenv_path()
    context = redirect_stdout(StringIO()) if quiet else nullcontext()
    with context:
        manager = CrossBuildEnvManager(xbuildenv_path)
        if manager.current_version is None:
            manager.install()

        manager.check_version_marker()

        return manager.pyodide_root


@functools.cache
def get_pyodide_root() -> Path:
    init_environment()
    return Path(os.environ["PYODIDE_ROOT"])


def search_pyodide_root(curdir: str | Path, *, max_depth: int = 10) -> Path | None:
    """
    Recursively search for the root of the Pyodide repository,
    by looking for the pyproject.toml file in the parent directories
    which contains the [tool._pyodide] section.
    """
    pyproject_path, pyproject_file = search_pyproject_toml(curdir, max_depth)

    if pyproject_path is None or pyproject_file is None:
        return None

    if "tool" in pyproject_file and "_pyodide" in pyproject_file["tool"]:
        return pyproject_path.parent

    return None


def in_xbuildenv() -> bool:
    pyodide_root = get_pyodide_root()
    return pyodide_root.name == "pyodide-root"


def _load_config_manager():
    """Lazily load ConfigManager so that we can avoid circular imports."""
    from pyodide_build.config import ConfigManager, CrossBuildEnvConfigManager

    return ConfigManager, CrossBuildEnvConfigManager


@functools.cache
def get_build_environment_vars(pyodide_root: Path) -> dict[str, str]:
    """
    Get common environment variables for the in-tree and out-of-tree build.
    """
    _, CrossBuildEnvConfigManager = _load_config_manager()

    config_manager = CrossBuildEnvConfigManager(pyodide_root)
    env = config_manager.to_env()

    env.update(
        {
            # This environment variable is used for packages to detect if they are built
            # for pyodide during build time
            "PYODIDE": "1",
            # This is the legacy environment variable used for the aforementioned purpose
            "PYODIDE_PACKAGE_ABI": "1",
            "PYTHONPATH": env["HOSTSITEPACKAGES"],
        }
    )

    return env


@functools.cache
def get_host_build_environment_vars() -> dict[str, str]:
    ConfigManager, _ = _load_config_manager()
    manager = ConfigManager()
    return manager.to_env()


def get_build_flag(name: str) -> str:
    """
    Get a value of a build flag.
    """
    pyodide_root = get_pyodide_root()
    build_vars = get_build_environment_vars(pyodide_root)
    if name not in build_vars:
        raise ValueError(f"Unknown build flag: {name}")

    return build_vars[name]


def get_host_build_flag(name: str) -> str:
    """
    Get a value of a build flag without accessing the cross-build environment.
    """
    build_vars = get_host_build_environment_vars()
    if name not in build_vars:
        raise ValueError(f"Unknown build flag: {name}")

    return build_vars[name]


def get_pyversion_major() -> str:
    return get_build_flag("PYMAJOR")


def get_pyversion_minor() -> str:
    return get_build_flag("PYMINOR")


def get_pyversion_major_minor() -> str:
    return f"{get_pyversion_major()}.{get_pyversion_minor()}"


def get_pyversion() -> str:
    return f"python{get_pyversion_major_minor()}"


def get_hostsitepackages() -> str:
    return get_build_flag("HOSTSITEPACKAGES")


@functools.cache
def get_unisolated_packages() -> list[str]:
    # TODO: Remove this function (and use remote package index)
    # https://github.com/pyodide/pyodide-build/issues/43
    PYODIDE_ROOT = get_pyodide_root()

    unisolated_file = PYODIDE_ROOT / "unisolated.txt"
    if unisolated_file.exists():
        # in xbuild env, read from file
        unisolated_packages = unisolated_file.read_text().splitlines()
    else:
        from pyodide_build.recipe.loader import load_all_recipes

        unisolated_packages = []
        recipe_dir = PYODIDE_ROOT / "packages"
        recipes = load_all_recipes(recipe_dir)
        for name, config in recipes.items():
            if config.build.cross_build_env:
                unisolated_packages.append(name)

    return unisolated_packages


def platform() -> str:
    emscripten_version = get_build_flag("PYODIDE_EMSCRIPTEN_VERSION")
    version = emscripten_version.replace(".", "_")
    return f"emscripten_{version}_wasm32"


def wheel_platform() -> str:
    abi_version = get_build_flag("PYODIDE_ABI_VERSION")
    return f"pyodide_{abi_version}_wasm32"


def pyodide_tags_() -> Iterator[Tag]:
    """
    Returns the sequence of tag triples for the Pyodide interpreter.

    The sequence is ordered in decreasing specificity.
    """
    PYMAJOR = get_pyversion_major()
    PYMINOR = get_pyversion_minor()
    PLATFORMS = [platform(), wheel_platform()]
    python_version = (int(PYMAJOR), int(PYMINOR))
    yield from cpython_tags(platforms=PLATFORMS, python_version=python_version)
    yield from compatible_tags(platforms=PLATFORMS, python_version=python_version)
    # Following line can be removed once packaging 22.0 is released and we update to it.
    yield Tag(interpreter=f"cp{PYMAJOR}{PYMINOR}", abi="none", platform="any")


@functools.cache
def pyodide_tags() -> Sequence[Tag]:
    return list(pyodide_tags_())


def replace_so_abi_tags(wheel_dir: Path) -> None:
    """Replace native abi tag with emscripten abi tag in .so file names"""
    import sysconfig

    build_soabi = sysconfig.get_config_var("SOABI")
    assert build_soabi
    ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
    assert ext_suffix
    build_triplet = "-".join(build_soabi.split("-")[2:])
    host_triplet = get_build_flag("PLATFORM_TRIPLET")
    for file in wheel_dir.glob(f"**/*{ext_suffix}"):
        file.rename(file.with_name(file.name.replace(build_triplet, host_triplet)))


def emscripten_version() -> str:
    return get_build_flag("PYODIDE_EMSCRIPTEN_VERSION")


def get_emscripten_version_info() -> str:
    """Extracted for testing purposes."""
    return subprocess.run(
        ["emcc", "-v"], capture_output=True, encoding="utf8", check=True
    ).stderr


def check_emscripten_version() -> None:
    skip = get_build_flag("SKIP_EMSCRIPTEN_VERSION_CHECK")
    if to_bool(skip):
        return

    needed_version = emscripten_version()
    try:
        version_info = get_emscripten_version_info()
    except FileNotFoundError:
        raise RuntimeError(
            f"No Emscripten compiler found. Need Emscripten version {needed_version}"
        ) from None
    installed_version = None
    try:
        for x in reversed(version_info.partition("\n")[0].split(" ")):
            # (X.Y.Z) or (X.Y.Z)-git
            match = re.match(r"(\d+\.\d+\.\d+)(-\w+)?", x)
            if match:
                installed_version = match.group(1)
                break
    except Exception:
        raise RuntimeError("Failed to determine Emscripten version.") from None
    if installed_version is None:
        raise RuntimeError("Failed to determine Emscripten version.")
    if installed_version != needed_version:
        raise RuntimeError(
            f"Incorrect Emscripten version {installed_version}. Need Emscripten version {needed_version}"
        )


def local_versions() -> dict[str, str]:
    """
    Returns the versions of the local Python interpreter and the pyodide-build.
    This information is used for checking compatibility with the cross-build environment.
    """
    return {
        "python": f"{sys.version_info.major}.{sys.version_info.minor}",
        "pyodide-build": __version__,
        # "emscripten": "TODO"
    }


def _create_constraints_file() -> str:
    try:
        constraints = get_build_flag("PIP_CONSTRAINT")
    except ValueError:
        return ""

    if not constraints:
        return ""

    if len(constraints.split(maxsplit=1)) > 1:
        raise ValueError(
            "PIP_CONSTRAINT contains spaces so pip will misinterpret it. Make sure the path to pyodide has no spaces.\n"
            "See https://github.com/pypa/pip/issues/13283"
        )

    constraints_file = Path(constraints)
    if not constraints_file.is_file():
        constraints_file.parent.mkdir(parents=True, exist_ok=True)
        constraints_file.write_text("")
    return constraints



#!/usr/bin/env python3
"""Helper for cross-compiling Python binary extensions.

Python has never had a proper cross-compilation story. This is a hack, which
miraculously works, to get around that.
The gist is we compile the package replacing calls to the compiler and linker
with wrappers that adjusting include paths and flags as necessary for
cross-compiling and then pass the command long to emscripten.
"""

import json
import os
import sys
from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import Literal, NamedTuple

from __main__ import __file__ as INVOKED_PATH_STR

INVOKED_PATH = Path(INVOKED_PATH_STR)

SYMLINKS = {
    "cc",
    "c++",
    "ld",
    "lld",
    "ar",
    "gcc",
    "ranlib",
    "strip",
    "gfortran",
    "cargo",
    "cmake",
    "meson",
    "install_name_tool",
    "otool",
}
IS_COMPILER_INVOCATION = INVOKED_PATH.name in SYMLINKS

if IS_COMPILER_INVOCATION:
    # If possible load from environment variable, if necessary load from disk.
    if "PYWASMCROSS_ARGS" in os.environ:
        PYWASMCROSS_ARGS = json.loads(os.environ["PYWASMCROSS_ARGS"])
    else:
        try:
            with open(INVOKED_PATH.parent / "pywasmcross_env.json") as f:
                PYWASMCROSS_ARGS = json.load(f)
        except FileNotFoundError:
            raise RuntimeError(
                "Invalid invocation: can't find PYWASMCROSS_ARGS."
                f" Invoked from {INVOKED_PATH}."
            ) from None

    sys.path = PYWASMCROSS_ARGS.pop("PYTHONPATH")
    os.environ["PATH"] = (
        os.environ["BUILD_ENV_SCRIPTS_DIR"] + ":" + PYWASMCROSS_ARGS.pop("PATH")
    )
    # restore __name__ so that relative imports work as we expect
    __name__ = PYWASMCROSS_ARGS.pop("orig__name__")


class CrossCompileArgs(NamedTuple):
    """
    Arguments for cross-compiling a package.
    """

    cflags: str = ""
    cxxflags: str = ""
    ldflags: str = ""

    # The name of the package being compiled
    # This is used to apply package-specific fixes, such as scipy
    pkgname: str = ""
    target_install_dir: str = ""  # The path to the target Python installation
    pythoninclude: str = ""  # path to the cross-compiled Python include directory
    exports: Literal["whole_archive", "requested", "pyinit"] | list[str] = "pyinit"
    # Pyodide abi, e.g., 2025_0
    # Sometimes we have to inject compile flags only for certain abis.
    abi: str = ""


def is_link_cmd(line: list[str]) -> bool:
    """
    Check if the command is a linker invocation.
    """
    for arg in line:
        if not arg or arg[0] == "-":
            continue

        if arg.endswith(".so"):
            return True

        dot_so_pos = arg.find(".so.")
        if dot_so_pos != -1:
            version = arg[dot_so_pos + 4 :]
            # versioned shared libraries start with a digit, and consist of only digits and dots.
            if version and version[0].isdigit():
                if all(c.isdigit() or c == "." for c in version):
                    return True

    return False


def replay_genargs_handle_dashl(arg: str, used_libs: set[str], abi: str) -> str | None:
    """
    Figure out how to replace a `-lsomelib` argument.

    Parameters
    ----------
    arg
        The argument we are replacing. Must start with `-l`.

    used_libs
        The libraries we've used so far in this command. emcc fails out if `-lsomelib`
        occurs twice, so we have to track this.

    Returns
    -------
        The new argument, or None to delete the argument.
    """
    assert arg.startswith("-l")

    if arg in {"-lffi", "-lgfortran"}:
        return None

    # Some Emscripten libraries that use setjmp/longjmp.
    # The Emscripten linker should automatically know to use these variants so
    # this shouldn't be necessary.
    if abi > "2025" and arg in ("-lfreetype", "-lpng"):
        arg += "-legacysjlj"

    # WASM link doesn't like libraries being included twice
    # skip second one
    if arg in used_libs:
        return None
    used_libs.add(arg)
    return arg


def replay_genargs_handle_dashI(arg: str, target_install_dir: str) -> str | None:
    """
    Figure out how to replace a `-Iincludepath` argument.

    Parameters
    ----------
    arg
        The argument we are replacing. Must start with `-I`.

    target_install_dir
        The target_install_dir argument.

    Returns
    -------
        The new argument, or None to delete the argument.
    """
    assert arg.startswith("-I")

    include_path_str = arg[2:]

    # Don't include any system directories
    if include_path_str.startswith("/usr"):
        return None

    # Replace local Python include paths with the cross compiled ones
    include_path = str(Path(include_path_str).resolve())

    if include_path.startswith(sys.prefix + "/include/python"):
        return arg.replace("-I" + sys.prefix, "-I" + target_install_dir)

    if include_path.startswith(sys.base_prefix + "/include/python"):
        return arg.replace("-I" + sys.base_prefix, "-I" + target_install_dir)

    return arg


def replay_genargs_handle_linker_opts(arg: str) -> str | None:
    """
    ignore some link flags
    it should not check if `arg == "-Wl,-xxx"` and ignore directly here,
    because arg may be something like "-Wl,-xxx,-yyy" where we only want
    to ignore "-xxx" but not "-yyy".
    """
    assert arg.startswith("-Wl")

    excluded_linker_opts = (
        "-Bsymbolic-functions",
        # breaks emscripten see https://github.com/emscripten-core/emscripten/issues/14460
        "--strip-all",
        "-strip-all",
        # wasm-ld does not recognize some link flags
        "--sort-common",
        "--as-needed",
        # macOS-specific linker flags that wasm-ld doesn't understand
        "-headerpad_max_install_names",
        "-dead_strip_dylibs",
    )

    excluded_linker_prefixes = (
        "--sysroot=",  # ignore unsupported --sysroot compile argument used in conda
        "--version-script=",
        "-R/",  # wasm-ld does not accept -R (runtime libraries)
        "-R.",  # wasm-ld does not accept -R (runtime libraries)
        "--exclude-libs=",
    )

    link_opts = arg.split(",")[1:]
    new_link_opts = ["-Wl"]

    for opt in link_opts:
        if opt in excluded_linker_opts:
            continue

        if opt.startswith(excluded_linker_prefixes):
            continue

        new_link_opts.append(opt)

    if len(new_link_opts) > 1:
        return ",".join(new_link_opts)
    else:
        return None


def replay_genargs_handle_argument(arg: str) -> str | None:
    """
    Figure out how to replace a general argument.

    Parameters
    ----------
    arg
        The argument we are replacing. Must not start with `-I` or `-l`.

    Returns
    -------
        The new argument, or None to delete the argument.
    """
    assert not arg.startswith("-I")  # should be handled by other functions
    assert not arg.startswith("-l")
    assert not arg.startswith("-Wl,")

    # Don't include any system directories
    if arg.startswith("-L/usr"):
        return None

    # fmt: off
    excluded_arguments = (
        # threading is disabled for now
        "-pthread",
        # this only applies to compiling fortran code, but we already f2c'd
        "-ffixed-form",
        "-fallow-argument-mismatch",
        # On Mac, we need to omit some darwin-specific arguments
        "-bundle", "-undefined", "dynamic_lookup",
        # This flag is needed to build numpy with SIMD optimization which we currently disable
        "-mpopcnt",
        # gcc flag that clang does not support
        "-Bsymbolic-functions",
        '-fno-second-underscore',
        '-fstack-protector',  # doesn't work?
        '-fno-strict-overflow',  # warning: argument unused during compilation
        "-mno-sse2", # warning: argument unused during compilation
        "-mno-avx2", # warning: argument unused during compilation
        "-std=legacy", # fortran flag that clang does not support
    )
    # fmt: on

    if arg in excluded_arguments:
        return None

    if arg.startswith("-J"):  # fortran flag that clang does not support
        return None

    return arg


def get_cmake_compiler_flags() -> list[str]:
    """
    Generate cmake compiler flags.
    emcmake will set these values to emcc, em++, ...
    but we need to set them to cc, c++, in order to make them pass to pywasmcross.
    Returns
    -------
    The commandline flags to pass to cmake.
    """
    compiler_flags = {
        "CMAKE_C_COMPILER": "cc",
        "CMAKE_CXX_COMPILER": "c++",
        "CMAKE_AR": "ar",
        "CMAKE_C_COMPILER_AR": "ar",
        "CMAKE_CXX_COMPILER_AR": "ar",
    }

    flags = []
    symlinks_dir = Path(sys.argv[0]).parent
    for key, value in compiler_flags.items():
        assert value in SYMLINKS

        flags.append(f"-D{key}={symlinks_dir / value}")

    return flags


def _calculate_object_exports_readobj_parse(output: str) -> list[str]:
    """
    >>> _calculate_object_exports_readobj_parse(
    ...     '''
    ...     Format: WASM \\n Arch: wasm32 \\n AddressSize: 32bit
    ...     Sections [
    ...         Section { \\n Type: TYPE (0x1)   \\n Size: 5  \\n Offset: 8  \\n }
    ...         Section { \\n Type: IMPORT (0x2) \\n Size: 32 \\n Offset: 19 \\n }
    ...     ]
    ...     Symbol {
    ...         Name: g2 \\n Type: FUNCTION (0x0) \\n
    ...         Flags [ (0x0) \\n ]
    ...         ElementIndex: 0x2
    ...     }
    ...     Symbol {
    ...         Name: f2 \\n Type: FUNCTION (0x0) \\n
    ...         Flags [ (0x4) \\n VISIBILITY_HIDDEN (0x4) \\n ]
    ...         ElementIndex: 0x1
    ...     }
    ...     Symbol {
    ...         Name: l  \\n Type: FUNCTION (0x0)
    ...         Flags [ (0x10)\\n UNDEFINED (0x10) \\n ]
    ...         ImportModule: env
    ...         ElementIndex: 0x0
    ...     }
    ...     '''
    ... )
    ['g2']
    """
    result = []
    insymbol = False
    for line in output.split("\n"):
        line = line.strip()
        if line == "Symbol {":
            insymbol = True
            export = True
            name = None
            symbol_lines = [line]
            continue
        if not insymbol:
            continue
        symbol_lines.append(line)
        if line.startswith("Name:"):
            name = line.removeprefix("Name:").strip()
        if line.startswith(("BINDING_LOCAL", "UNDEFINED", "VISIBILITY_HIDDEN")):
            export = False
        if line == "}":
            insymbol = False
            if export:
                if not name:
                    raise RuntimeError(
                        "Didn't find symbol's name:\n" + "\n".join(symbol_lines)
                    )
                result.append(name)
    return result


def calculate_object_exports_readobj(objects: list[str]) -> list[str] | None:
    from shutil import which
    from subprocess import run as sp_run

    # This works for bootstrapped Emscripten via GitHub sources.
    # llvm-readobj might not be available this way with Homebrew
    # or conda-forge distributions of Emscripten.
    which_emcc = which("emcc")
    assert which_emcc
    emcc = Path(which_emcc)
    readobj = (emcc / "../../bin/llvm-readobj").resolve()
    if readobj.exists():
        readobj_path = str(readobj)
    else:
        readobj_path = which("llvm-readobj")
    if not readobj_path:
        print("Failed to find llvm-readobj, quitting", file=sys.stdout)
        sys.exit(1)

    args = [
        readobj_path,
        "--section-details",
        "-st",
    ] + objects
    completedprocess = sp_run(
        args,
        encoding="utf8",
        capture_output=True,
        env={"PATH": os.environ["PATH"]},
        check=False,
    )
    if completedprocess.returncode:
        print(f"Command '{' '.join(args)}' failed. Output to stderr was:")
        print(completedprocess.stderr)
        sys.exit(completedprocess.returncode)

    if "bitcode files are not supported" in completedprocess.stderr:
        return None

    return _calculate_object_exports_readobj_parse(completedprocess.stdout)


def calculate_object_exports_nm(objects: list[str]) -> list[str]:
    from subprocess import run as sp_run

    args = ["emnm", "-j", "--export-symbols"] + objects
    result = sp_run(
        args,
        encoding="utf8",
        capture_output=True,
        env={"PATH": os.environ["PATH"]},
        check=False,
    )
    if result.returncode:
        print(f"Command '{' '.join(args)}' failed. Output to stderr was:")
        print(result.stderr)
        sys.exit(result.returncode)
    return result.stdout.splitlines()


def filter_objects(line: list[str]) -> list[str]:
    """
    Collect up all the object files and archive files being linked.
    """
    return [
        arg
        for arg in line
        if arg.endswith((".a", ".o"))
        or arg.startswith(
            "@"
        )  # response file (https://gcc.gnu.org/wiki/Response_Files)
    ]


def calculate_exports(line: list[str], export_all: bool) -> Iterable[str]:
    """
    List out symbols from object files and archive files that are marked as public.
    If ``export_all`` is ``True``, then return all public symbols.
    If not, return only the public symbols that begin with `PyInit`.
    """
    objects = filter_objects(line)
    exports = None
    # Using emnm is simpler but it cannot handle bitcode. If we're only
    # exporting the PyInit symbols, save effort by using nm.
    if export_all:
        exports = calculate_object_exports_readobj(objects)
    if exports is None:
        # Either export_all is false or we are linking at least one bitcode
        # object. Fall back to a more conservative estimate of the symbols
        # exported. This can export things with `__visibility__("hidden")`
        exports = calculate_object_exports_nm(objects)
    if export_all:
        return exports
    return (x for x in exports if x.startswith("PyInit"))


def get_export_flags(
    line: list[str],
    exports: Literal["whole_archive", "requested", "pyinit"] | list[str],
) -> Iterator[str]:
    """
    If "whole_archive" was requested, no action is needed. Otherwise, add
    `-sSIDE_MODULE=2` and the appropriate export list.
    """
    from tempfile import NamedTemporaryFile

    if exports == "whole_archive":
        return
    yield "-sSIDE_MODULE=2"
    if isinstance(exports, str):
        export_list = calculate_exports(line, exports == "requested")
    else:
        export_list = exports

    prefixed_exports = ["_" + x for x in export_list]

    with NamedTemporaryFile(mode="w", delete=False) as f:
        # Use a response file to avoid command line length limits
        f.write(json.dumps(prefixed_exports))

    yield f"-sEXPORTED_FUNCTIONS=@{f.name}"


def handle_command_generate_args(  # noqa: C901
    line: list[str], build_args: CrossCompileArgs
) -> list[str]:
    """
    A helper command for `handle_command` that generates the new arguments for
    the compilation.

    Unlike `handle_command` this avoids I/O: it doesn't sys.exit, it doesn't run
    subprocesses, it doesn't create any files, and it doesn't write to stdout.

    Parameters
    ----------
    line The original compilation command as a list e.g., ["gcc", "-c",
        "input.c", "-o", "output.c"]

    build_args The arguments that pywasmcross was invoked with

    Returns
    -------
        An updated argument list suitable for use with emscripten.

    Examples
    --------

    >>> from collections import namedtuple
    >>> Args = namedtuple('args', ['cflags', 'cxxflags', 'ldflags', 'target_install_dir'])
    >>> args = Args(cflags='', cxxflags='', ldflags='', target_install_dir='')
    >>> handle_command_generate_args(['gcc', 'test.c'], args)
    ['emcc', 'test.c', '-Werror=implicit-function-declaration', '-Werror=mismatched-parameter-types', '-Werror=return-type']
    """
    if "-print-multiarch" in line:
        return ["echo", "wasm32-emscripten"]
    if len(line) == 2 and line[1] == "-v":
        return ["emcc", "-v"]

    cmd = line[0]
    if cmd in {"c++", "g++"}:
        new_args = ["em++"]
    elif cmd in {"cc", "gcc", "ld", "lld"}:
        new_args = ["emcc"]
        # distutils doesn't use the c++ compiler when compiling c++ <sigh>
        if any(arg.endswith((".cpp", ".cc")) for arg in line):
            new_args = ["em++"]
    elif cmd == "ar":
        line[0] = "emar"
        return line
    elif cmd == "cmake":
        # If it is a build/install command, or running a script, we don't do anything.
        if "--build" in line or "--install" in line or "-P" in line or "-E" in line:
            return line

        flags = get_cmake_compiler_flags()
        line[:1] = [
            "emcmake",
            "cmake",
            *flags,
        ]
        return line
    elif cmd == "meson":
        if line[:2] != ["meson", "setup"]:
            return line

        if "MESON_CROSS_FILE" in os.environ:
            line[:2] = [
                "meson",
                "setup",
                "--cross-file",
                os.environ["MESON_CROSS_FILE"],
            ]

        return line
    elif cmd in {"install_name_tool", "otool"}:
        # In MacOS, meson tries to run install_name_tool to fix the rpath of the shared library
        # assuming that it is an ELF file. We need to skip this step.
        # See: https://github.com/mesonbuild/meson/issues/8027
        return ["echo", *line]
    elif cmd == "ranlib":
        line[0] = "emranlib"
        return line
    elif cmd == "strip":
        line[0] = "emstrip"
        return line
    else:
        return line

    used_libs: set[str] = set()
    # Go through and adjust arguments
    for arg in line[1:]:
        if new_args[-1].startswith("-B") and "compiler_compat" in arg:
            # conda uses custom compiler search paths with the compiler_compat folder.
            # Ignore it.
            del new_args[-1]
            continue

        if arg.startswith("-l"):
            result = replay_genargs_handle_dashl(arg, used_libs, build_args.abi)
        elif arg.startswith("-I"):
            result = replay_genargs_handle_dashI(arg, build_args.target_install_dir)
        elif arg.startswith("-Wl"):
            result = replay_genargs_handle_linker_opts(arg)
        else:
            result = replay_genargs_handle_argument(arg)

        if result:
            new_args.append(result)

    new_args.extend(
        [
            "-Werror=implicit-function-declaration",
            "-Werror=mismatched-parameter-types",
            "-Werror=return-type",
        ]
    )

    # set linker and C flags to error on anything to do with function declarations being wrong.
    # Better to fail at compile or link time.
    if is_link_cmd(line):
        new_args.append("-Wl,--fatal-warnings")
        new_args.extend(build_args.ldflags.split())
        new_args.extend(get_export_flags(line, build_args.exports))

    if "-c" in line:
        if new_args[0] == "emcc":
            new_args.extend(build_args.cflags.split())
        elif new_args[0] == "em++":
            new_args.extend(build_args.cflags.split() + build_args.cxxflags.split())

        if build_args.pythoninclude:
            new_args.extend(["-I", build_args.pythoninclude])

    return new_args


def handle_command(
    line: list[str],
    build_args: CrossCompileArgs,
) -> int:
    """Handle a compilation command. Exit with an appropriate exit code when done.

    Parameters
    ----------
    line : iterable
       an iterable with the compilation arguments
    build_args : BuildArgs
       a container with additional compilation options
    """
    from subprocess import run as sp_run

    if line[0] == "gfortran":
        from _f2c_fixes import replay_f2c

        tmp = replay_f2c(line)
        if tmp is None:
            # No source file, it's a query for information about the compiler. Pretend we're
            # gfortran by letting gfortran handle it
            return sp_run(line, check=False).returncode

        line = tmp

    new_args = handle_command_generate_args(line, build_args)

    if build_args.pkgname == "scipy":
        from _f2c_fixes import scipy_fixes

        scipy_fixes(new_args)

    result = sp_run(new_args, check=False)
    return result.returncode


def compiler_main():
    build_args = CrossCompileArgs(
        pkgname=PYWASMCROSS_ARGS["pkgname"],
        cflags=PYWASMCROSS_ARGS["cflags"],
        cxxflags=PYWASMCROSS_ARGS["cxxflags"],
        ldflags=PYWASMCROSS_ARGS["ldflags"],
        target_install_dir=PYWASMCROSS_ARGS["target_install_dir"],
        pythoninclude=PYWASMCROSS_ARGS["pythoninclude"],
        exports=PYWASMCROSS_ARGS["exports"],
        abi=PYWASMCROSS_ARGS["abi"],
    )
    basename = Path(sys.argv[0]).name
    args = list(sys.argv)
    args[0] = basename
    sys.exit(handle_command(args, build_args))


if IS_COMPILER_INVOCATION:
    compiler_main()



# Class for generating "views", i.e., tabular and JSON outputs from
# metadata objects, currently used in the xbuildenv CLI (search command).


import json
from dataclasses import dataclass


@dataclass
class MetadataView:
    version: str
    python: str
    emscripten: str
    pyodide_build: dict[str, str | None]
    compatible: bool

    @classmethod
    def to_table(cls, views: list["MetadataView"]) -> str:
        columns = [
            ("Version", 10),
            ("Python", 10),
            ("Emscripten", 10),
            ("pyodide-build", 25),
            ("Compatible", 10),
        ]

        # Unicode box-drawing characters
        top_left, top_right = "┌", "┐"
        bottom_left, bottom_right = "└", "┘"
        horizontal, vertical = "─", "│"
        t_down, t_up, t_right, t_left = "┬", "┴", "├", "┤"
        cross = "┼"

        # Table elements
        top_border = (
            top_left
            + t_down.join(horizontal * (width + 2) for _, width in columns)
            + top_right
        )
        header = (
            vertical
            + vertical.join(f" {name:<{width}} " for name, width in columns)
            + vertical
        )
        separator = (
            t_right
            + cross.join(horizontal * (width + 2) for _, width in columns)
            + t_left
        )
        bottom_border = (
            bottom_left
            + t_up.join(horizontal * (width + 2) for _, width in columns)
            + bottom_right
        )

        ### Printing
        table = [top_border, header, separator]
        for view in views:
            pyodide_build_range = (
                f"{view.pyodide_build['min'] or ''} - {view.pyodide_build['max'] or ''}"
            )
            row = [
                f"{view.version:<{columns[0][1]}}",
                f"{view.python:<{columns[1][1]}}",
                f"{view.emscripten:<{columns[2][1]}}",
                f"{pyodide_build_range:<{columns[3][1]}}",
                f"{'Yes' if view.compatible else 'No':<{columns[4][1]}}",
            ]
            table.append(
                vertical + vertical.join(f" {cell} " for cell in row) + vertical
            )
        table.append(bottom_border)
        return "\n".join(table)

    @classmethod
    def to_json(cls, views: list["MetadataView"]) -> str:
        result = json.dumps(
            {
                "environments": [
                    {
                        "version": view.version,
                        "python": view.python,
                        "emscripten": view.emscripten,
                        "pyodide_build": view.pyodide_build,
                        "compatible": view.compatible,
                    }
                    for view in views
                ]
            },
            indent=2,
        )
        return result



from pathlib import Path

import typer

from pyodide_build.build_env import local_versions
from pyodide_build.common import default_xbuildenv_path
from pyodide_build.views import MetadataView
from pyodide_build.xbuildenv import CrossBuildEnvManager
from pyodide_build.xbuildenv_releases import (
    cross_build_env_metadata_url,
    load_cross_build_env_metadata,
)

DEFAULT_PATH = default_xbuildenv_path()

app = typer.Typer(no_args_is_help=True)


@app.callback()
def callback():
    """
    Manage cross-build environment for building packages for Pyodide.
    """


def check_xbuildenv_root(path: Path) -> None:
    if not path.is_dir():
        typer.echo(f"Cross-build environment not found in {path.resolve()}")
        raise typer.Exit(1)


@app.command("install")
def _install(
    version: str = typer.Argument(
        None, help="version of cross-build environment to install"
    ),
    path: Path = typer.Option(
        DEFAULT_PATH,
        envvar="PYODIDE_XBUILDENV_PATH",
        help="destination to download cross-build environment directory to.",
    ),
    url: str = typer.Option(None, help="URL to download cross-build environment from"),
    force_install: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="force installation even if the version is not compatible",
    ),
) -> None:
    """
    Install cross-build environment.

    The installed environment is the same as the one that would result from
    `PYODIDE_PACKAGES='scipy' make` except that it is much faster.
    The goal is to enable out-of-tree builds for binary packages that depend
    on numpy or scipy.
    """
    manager = CrossBuildEnvManager(path)

    if url:
        manager.install(url=url, force_install=force_install)
    else:
        manager.install(version=version, force_install=force_install)

    typer.echo(f"Pyodide cross-build environment installed at {path.resolve()}")


@app.command("version")
def _version(
    path: Path = typer.Option(
        DEFAULT_PATH, help="path to cross-build environment directory"
    ),
) -> None:
    """
    Print current version of cross-build environment.
    """
    check_xbuildenv_root(path)
    manager = CrossBuildEnvManager(path)
    version = manager.current_version
    if not version:
        typer.echo("No version selected")
        raise typer.Exit(1)
    else:
        typer.echo(version)


@app.command("versions")
def _versions(
    path: Path = typer.Option(
        DEFAULT_PATH, help="path to cross-build environment directory"
    ),
) -> None:
    """
    Print all installed versions of cross-build environment.
    """
    check_xbuildenv_root(path)
    manager = CrossBuildEnvManager(path)
    versions = manager.list_versions()
    current_version = manager.current_version

    for version in versions:
        if version == current_version:
            typer.echo(f"* {version}")
        else:
            typer.echo(f"  {version}")


@app.command("uninstall")
def _uninstall(
    version: str = typer.Argument(
        None, help="version of cross-build environment to uninstall"
    ),
    path: Path = typer.Option(
        DEFAULT_PATH, help="path to cross-build environment directory"
    ),
) -> None:
    """
    Uninstall cross-build environment.
    """
    check_xbuildenv_root(path)
    manager = CrossBuildEnvManager(path)
    v = manager.uninstall_version(version)
    typer.echo(f"Pyodide cross-build environment {v} uninstalled")


@app.command("use")
def _use(
    version: str = typer.Argument(
        ..., help="version of cross-build environment to use"
    ),
    path: Path = typer.Option(
        DEFAULT_PATH, help="path to cross-build environment directory"
    ),
) -> None:
    """
    Select a version of cross-build environment to use.
    """
    check_xbuildenv_root(path)
    manager = CrossBuildEnvManager(path)
    manager.use_version(version)
    typer.echo(f"Pyodide cross-build environment {version} is now in use")


@app.command("search")
def _search(
    metadata_path: str = typer.Option(
        None,
        "--metadata",
        help="path to cross-build environment metadata file. It can be a URL or a local file. If not given, the default metadata file is used.",
    ),
    show_all: bool = typer.Option(
        False,
        "--all",
        "-a",
        help="search all versions, without filtering out incompatible ones",
    ),
    json_output: bool = typer.Option(
        False,
        "--json",
        help="output results in JSON format",
    ),
) -> None:
    """
    Search for available versions of cross-build environment.
    """

    # TODO: cache the metadata file somewhere to avoid downloading it every time
    metadata_path = metadata_path or cross_build_env_metadata_url()
    metadata = load_cross_build_env_metadata(metadata_path)
    local = local_versions()

    if show_all:
        releases = metadata.list_compatible_releases()
    else:
        releases = metadata.list_compatible_releases(
            python_version=local["python"],
            pyodide_build_version=local["pyodide-build"],
        )

    if not releases:
        typer.echo(
            "No compatible cross-build environment found for your system. Try using --all to see all versions."
        )
        raise typer.Exit(1)

    # Generate views for the metadata objects (currently tabular or JSON)
    views = [
        MetadataView(
            version=release.version,
            python=release.python_version,
            emscripten=release.emscripten_version,
            pyodide_build={
                "min": release.min_pyodide_build_version,
                "max": release.max_pyodide_build_version,
            },
            compatible=release.is_compatible(
                python_version=local["python"],
                pyodide_build_version=local["pyodide-build"],
            ),
        )
        for release in releases
    ]

    if json_output:
        print(MetadataView.to_json(views))
    else:
        print(MetadataView.to_table(views))



from pathlib import Path
from textwrap import dedent
from urllib.parse import urlparse

from pyodide_lock.spec import PackageSpec


def create_package_index(
    packages: dict[str, PackageSpec], target_dir: Path, dist_url: str
) -> None:
    """Create a pip-compatible Python package index to be used with a Pyodide virtual
    environment.

    To use, pass as an `--index-url` or `--extra-index-url` parameter to pip.
    The argument should be a `file:` url pointing to the `package_index` folder (or
    if you serve `package_index` it can be a normal url). It is also used
    automatically by Pyodide virtual environments created from a release version
    of Pyodide.

    Parameters
    ----------
    packages:
        A dictionary of packages that we want to index. This should be the
        "packages" field from pyodide-lock.json.

    target_dir:
        Where to put the  index. It will be placed in a subfolder of
        target_dir called `package_index`. `target_dir` should exist but
        `target_dir/package_index` should not exist.

    dist_url:
        The CDN url to download packages from. This will be hard coded into the
        generated index. If you wish to install from local files, then prefix
        with `file:` e.g., `f"file:{pyodide_root}/dist"`.
    """
    # We only want to index the wheels
    packages = {
        pkgname: pkginfo
        for (pkgname, pkginfo) in packages.items()
        if pkginfo.file_name.endswith(".whl")
    }
    if not target_dir.exists():
        raise RuntimeError(f"target_dir={target_dir} does not exist")

    index_dir = target_dir / "package_index"
    if index_dir.exists():
        raise RuntimeError(f"{index_dir} already exists")
    index_dir.mkdir()

    # Create top level index
    packages_str = "\n".join(f'<a href="{x}/">{x}</a>' for x in packages.keys())
    index_html = dedent(
        f"""
        <!DOCTYPE html>
        <html>
        <head>
        <meta name="pypi:repository-version" content="1.0">
        <title>Simple index</title>
        </head>
        <body>
        {packages_str}
        </body>
        </html>
        """
    ).strip()

    (index_dir / "index.html").write_text(index_html)

    files_template = dedent(
        """
        <!DOCTYPE html>
        <html>
        <head>
        <meta name="pypi:repository-version" content="1.0">
        <title>Links for {pkgname}</title>
        </head>
        <body>
        <h1>Links for {pkgname}</h1>
        {links}
        </body>
        </html>
        """
    ).strip()

    for pkgname, pkginfo in packages.items():
        pkgdir = index_dir / pkgname
        filename = pkginfo.file_name
        if urlparse(filename).scheme:
            href = filename
        else:
            shasum = pkginfo.sha256
            href = f"{dist_url}{filename}#sha256={shasum}"
        links_str = f'<a href="{href}">{pkgname}</a>\n'
        files_html = files_template.format(pkgname=pkgname, links=links_str)

        pkgdir.mkdir()
        (pkgdir / "index.html").write_text(files_html)



import functools
import os
import shutil


@functools.cache
def find_uv_bin() -> str | None:
    """
    Check if the uv executable is available.

    If the uv executable is available, return the path to the executable.
    Otherwise, return None.
    """
    try:
        import uv

        return uv.find_uv_bin()
    except (ModuleNotFoundError, FileNotFoundError):
        return shutil.which("uv")

    return None


def should_use_uv() -> bool:
    # UV environ is set to the uv executable path when the script is called with the uv executable.
    uv_environ = os.environ.get("UV")
    # double check by comparing the uv executable path with the one found by the uv package.
    return uv_environ and uv_environ == find_uv_bin()



import shutil
from collections.abc import Callable
from pathlib import Path
from tempfile import TemporaryDirectory

from pyodide_build._py_compile import _compile
from pyodide_build.common import make_zip_archive


def default_filterfunc(
    root: Path, excludes: list[str], stubs: list[str], verbose: bool = False
) -> Callable[[str, list[str]], set[str]]:
    """
    The default filter function used by `create_zipfile`.

    This function filters out several modules that are:

    - not supported in Pyodide due to browser limitations (e.g. `tkinter`)
    - unvendored from the standard library (e.g. `sqlite3`)
    """

    def _should_skip(path: Path) -> bool:
        """Skip common files that are not needed in the zip file."""
        name = path.name

        if path.is_dir() and name in ("__pycache__", "dist"):
            return True

        if path.is_dir() and name.endswith((".egg-info", ".dist-info")):
            return True

        if path.is_file() and name in (
            "LICENSE",
            "LICENSE.txt",
            "setup.py",
            ".gitignore",
        ):
            return True

        if path.is_file() and name.endswith(("pyi", "toml", "cfg", "md", "rst")):
            return True

        return False

    def filterfunc(path: Path | str, names: list[str]) -> set[str]:
        filtered_files = {(root / f).resolve() for f in excludes}

        # We have JS implementations of these modules, so we don't need to
        # include the Python ones. Checking the name of the root directory
        # is a bit of a hack, but it works...
        if root.name.startswith("python3"):
            filtered_files.update({root / f for f in stubs})

        path = Path(path).resolve()

        if _should_skip(path):
            return set(names)

        _names = []
        for name in names:
            fullpath = path / name

            if _should_skip(fullpath) or fullpath in filtered_files:
                if verbose:
                    print(f"Skipping {fullpath}")

                _names.append(name)

        return set(_names)

    return filterfunc


def create_zipfile(
    libdirs: list[Path],
    excludes: list[str] | None = None,
    stubs: list[str] | None = None,
    output: Path | str = "python",
    pycompile: bool = False,
    filterfunc: Callable[[str, list[str]], set[str]] | None = None,
    compression_level: int = 6,
) -> None:
    """
    Bundle Python standard libraries into a zip file.

    The basic idea of this function is similar to the standard library's
    {ref}`zipfile.PyZipFile` class.

    However, we need some additional functionality for Pyodide. For example:

    - We need to remove some unvendored modules, e.g. `sqlite3`
    - We need an option to "not" compile the files in the zip file

    hence this function.

    Parameters
    ----------
    libdirs
        List of paths to the directory containing the Python standard library or extra packages.

    excludes
        List of files to exclude from the zip file.

    stubs
        List of files that are replaced by JS implementations.

    output
        Path to the output zip file. Defaults to python.zip.

    pycompile
        Whether to compile the .py files into .pyc, by default False

    filterfunc
        A function that filters the files to be included in the zip file.
        This function will be passed to {ref}`shutil.copytree` 's ignore argument.
        By default, Pyodide's default filter function is used.

    compression_level
        Level of zip compression to apply. 0 means no compression. If a strictly
        positive integer is provided, ZIP_DEFLATED option is used.

    Returns
    -------
    BytesIO
        A BytesIO object containing the zip file.
    """

    archive = Path(output)
    excludes = excludes or []
    stubs = stubs or []

    with TemporaryDirectory() as temp_dir_str:
        temp_dir = Path(temp_dir_str)

        for libdir in libdirs:
            libdir = Path(libdir)

            if filterfunc is None:
                _filterfunc = default_filterfunc(libdir, excludes, stubs)

            shutil.copytree(libdir, temp_dir, ignore=_filterfunc, dirs_exist_ok=True)

        make_zip_archive(
            archive,
            temp_dir,
            compression_level=compression_level,
        )

    if pycompile:
        _compile(
            archive,
            archive,
            verbose=False,
            keep=False,
            compression_level=compression_level,
        )
